import copy

import pytest

from ase.build import molecule
from ase.calculators.emt import EMT
from ase.constraints import FixInternals
from ase.optimize.bfgs import BFGS


def setup_atoms():
    atoms = molecule('CH3CH2OH', vacuum=5.0)
    atoms.rattle(stdev=0.3)
    return atoms


def setup_fixinternals():
    atoms = setup_atoms()

    # Angles, Bonds, Dihedrals are built up with pairs of constraint
    # value and indices defining the constraint
    # Linear combinations of bond lengths are built up similarly with the
    # coefficients appended to the indices defining the constraint

    # Fix bond between atoms 1 and 2 to 1.4
    bond_def = [1, 2]
    target_bond = 1.4

    # Fix angle to whatever it was from the start
    angle_def = [6, 0, 1]
    target_angle = atoms.get_angle(*angle_def)

    # Fix this dihedral angle to whatever it was from the start
    dihedral_def = [6, 0, 1, 2]
    target_dihedral = atoms.get_dihedral(*dihedral_def)

    # Initialize constraint
    constr = FixInternals(bonds=[(target_bond, bond_def)],
                          angles_deg=[(target_angle, angle_def)],
                          dihedrals_deg=[(target_dihedral, dihedral_def)],
                          epsilon=1e-10)
    return (atoms, constr, bond_def, target_bond, angle_def, target_angle,
            dihedral_def, target_dihedral)


@pytest.mark.optimize()
def test_fixinternals():
    (atoms, constr, bond_def, target_bond, angle_def, _target_angle,
     dihedral_def, _target_dihedral) = setup_fixinternals()

    opt = BFGS(atoms)

    previous_angle = atoms.get_angle(*angle_def)
    previous_dihedral = atoms.get_dihedral(*dihedral_def)

    print('angle before', previous_angle)
    print('dihedral before', previous_dihedral)
    print('bond length before', atoms.get_distance(*bond_def))
    print('target bondlength', target_bond)

    atoms.calc = EMT()
    atoms.set_constraint(constr)
    print('-----Optimization-----')
    opt.run(fmax=0.01)

    new_angle = atoms.get_angle(*angle_def)
    new_dihedral = atoms.get_dihedral(*dihedral_def)
    new_bondlength = atoms.get_distance(*bond_def)

    print('angle after', new_angle)
    print('dihedral after', new_dihedral)
    print('bondlength after', new_bondlength)

    err1 = new_angle - previous_angle
    err2 = new_dihedral - previous_dihedral
    err3 = new_bondlength - target_bond

    print('error in angle', repr(err1))
    print('error in dihedral', repr(err2))
    print('error in bondlength', repr(err3))

    for err in [err1, err2, err3]:
        assert abs(err) < 1e-11


def setup_combos():
    atoms = setup_atoms()

    # Fix linear combination of two bond lengths with atom indices 0-8 and
    # 0-6 with weighting coefficients 1.0 and -1.0 to the current value.
    # In other words, fulfil the following constraint:
    # 1.0 * atoms.get_distance(2, 1) + -1.0 * atoms.get_distance(2, 3) = const.
    bondcombo_def = [[2, 1, 1.0], [2, 3, -1.0]]
    target_bondcombo = FixInternals.get_bondcombo(atoms, bondcombo_def)

    # Initialize constraint; 'None' value should be converted to current value
    constr = FixInternals(bondcombos=[(None, bondcombo_def)], epsilon=1e-10)
    return atoms, constr, bondcombo_def, target_bondcombo,


@pytest.mark.optimize()
def test_combos():
    atoms, constr, bondcombo_def, target_bondcombo = setup_combos()

    ref_bondcombo = FixInternals.get_bondcombo(atoms, bondcombo_def)

    atoms.calc = EMT()
    atoms.set_constraint(constr)

    atoms2 = atoms.copy()  # check if 'None' value converts to current value
    atoms2.set_positions(atoms2.get_positions())
    checked_bondcombo = False
    for subconstr in atoms2.constraints[0].constraints:
        if repr(subconstr).startswith('FixBondCombo'):
            assert subconstr.targetvalue == target_bondcombo
            checked_bondcombo = True
    assert checked_bondcombo

    opt = BFGS(atoms)
    opt.run(fmax=0.01)

    new_bondcombo = FixInternals.get_bondcombo(atoms, bondcombo_def)
    err_bondcombo = new_bondcombo - ref_bondcombo

    print('error in bondcombo:', repr(err_bondcombo))
    assert abs(err_bondcombo) < 1e-11


def test_index_shuffle():
    (atoms, constr, _bond_def, _target_bond, _angle_def, _target_angle,
     _dihedral_def, _target_dihedral) = setup_fixinternals()

    constr2 = copy.deepcopy(constr)

    # test no change, test constr.get_indices()
    assert all(a == b for a, b in zip(constr.get_indices(), (0, 1, 2, 6, 8)))
    constr.index_shuffle(atoms, range(len(atoms)))
    assert all(a == b for a, b in zip(constr.get_indices(), (0, 1, 2, 6, 8)))

    # test full constraint is not part of new slice
    with pytest.raises(IndexError):
        constr.index_shuffle(atoms, [0])

    # test correct shuffling
    constr2.index_shuffle(atoms, [1, 2, 0, 6])
    assert constr2.bonds[0][1] == [0, 1]
    assert constr2.angles[0][1] == [3, 2, 0]
    assert constr2.dihedrals[0][1] == [3, 2, 0, 1]


def test_combo_index_shuffle():
    atoms, constr, _bondcombo_def, _target_bondcombo = setup_combos()

    # test no change, test constr.get_indices()
    answer = (1, 2, 3)
    assert all(a == b for a, b in zip(constr.get_indices(), answer))
    constr.index_shuffle(atoms, range(len(atoms)))
    assert all(a == b for a, b in zip(constr.get_indices(), answer))


@pytest.mark.optimize()
def test_zero_distance_error():
    """Zero distances cannot be fixed due to a singularity in the derivative.
    """
    atoms = setup_atoms()
    constr = FixInternals(bonds=[(0.0, [1, 2])])
    atoms.calc = EMT()
    atoms.set_constraint(constr)
    opt = BFGS(atoms)
    with pytest.raises(ZeroDivisionError):
        for _ in opt.irun():
            atoms.get_distance(1, 2)


@pytest.mark.optimize()
def test_planar_angle_error():
    """Support for planar angles could be added in the future using
       dummy/ghost atoms. See issue #868."""
    atoms = setup_atoms()
    constr = FixInternals(angles_deg=[(180, [6, 0, 1])])
    atoms.calc = EMT()
    atoms.set_constraint(constr)
    opt = BFGS(atoms)
    with pytest.raises(ZeroDivisionError):
        opt.run()


@pytest.mark.optimize()
def test_undefined_dihedral_error():
    atoms = setup_atoms()
    pos = atoms.get_positions()
    pos[0:3] = [[8, 5, 5], [7, 5, 5], [6, 5, 5]]
    atoms.set_positions(pos)  # with undefined dihedral
    with pytest.raises(ZeroDivisionError):
        atoms.get_dihedral(6, 0, 1, 2)
    constr = FixInternals(dihedrals_deg=[(20., [6, 0, 1, 2])])
    atoms.calc = EMT()
    atoms.set_constraint(constr)
    opt = BFGS(atoms)
    with pytest.raises(ZeroDivisionError):
        opt.run()



import pytest


@pytest.mark.skip('CLI must support calculator profiles')
def test_abinit_cmdline(abinit_factory, cli):
    cli.shell("""
    ase build -x fcc -a 4.04 Al |
    ase -T run abinit -p xc=PBE,kpts=3.0,ecut=340,toldfe=1e-5,chksymbreak=0""",
              'abinit')



"""Test runner classes to run exciting simulations using subproces."""

import pytest

import ase.calculators.exciting.runner


@pytest.mark.parametrize(
    (
        "binary, run_argv, omp_num_threads,"), [
            ("exciting_serial", ['./'], 1),
            ("exciting_purempi", ['mpirun', '-np', '1'], 1),
            ("exciting_smp", ['./'], 1),
            ("exciting_mpismp", ['mpirun', '-np', '1'], 1)
    ])
def test_class_simple_binary_runner_init(
        tmpdir,
        binary,
        run_argv,
        omp_num_threads,
        excitingtools):
    """Test SimpleBinaryRunner."""
    binary = tmpdir / 'binary.exe'
    binary.write("Arbitrary text such that file exists")
    runner = ase.calculators.exciting.runner.SimpleBinaryRunner(
        binary=binary, run_argv=run_argv,
        omp_num_threads=omp_num_threads,
        directory=tmpdir,
        args=[])

    # Attributes
    assert runner.binary == binary, "Binary erroneously initialised"
    assert runner.run_argv == run_argv
    assert runner.omp_num_threads == omp_num_threads
    assert runner.directory == tmpdir
    assert runner.args == []


def test_compose_execution_list(tmpdir):
    """Test SimpleBinaryRunner's compose execution list method."""
    binary = tmpdir / 'binary.exe'
    binary.write("Arbitrary text such that file exists")
    runner = ase.calculators.exciting.runner.SimpleBinaryRunner(
        binary=binary, run_argv=['mpirun', '-np', '2'], omp_num_threads=1,
        directory=tmpdir,
        args=['input.txt'])
    # Attributes
    assert runner.binary == binary
    assert runner.run_argv == ['mpirun', '-np', '2']
    assert runner.omp_num_threads == 1
    assert runner.directory == tmpdir
    assert runner.args == ['input.txt']
    # Methods
    execute = runner.compose_execution_list()
    assert execute == ['mpirun', '-np', '2', str(binary), 'input.txt']



"""A collection of mutations that can be used."""
from math import cos, pi, sin

import numpy as np

from ase import Atoms
from ase.calculators.lammps.coordinatetransform import calc_rotated_cell
from ase.cell import Cell
from ase.ga.offspring_creator import CombinationMutation, OffspringCreator
from ase.ga.utilities import (
    atoms_too_close,
    atoms_too_close_two_sets,
    gather_atoms_by_tag,
    get_rotation_matrix,
)


class RattleMutation(OffspringCreator):
    """An implementation of the rattle mutation as described in:

    R.L. Johnston Dalton Transactions, Vol. 22,
    No. 22. (2003), pp. 4193-4207

    Parameters:

    blmin: Dictionary defining the minimum distance between atoms
        after the rattle.

    n_top: Number of atoms optimized by the GA.

    rattle_strength: Strength with which the atoms are moved.

    rattle_prop: The probability with which each atom is rattled.

    test_dist_to_slab: whether to also make sure that the distances
        between the atoms and the slab satisfy the blmin.

    use_tags: if True, the atomic tags will be used to preserve
        molecular identity. Same-tag atoms will then be
        displaced collectively, so that the internal
        geometry is preserved.

    rng: Random number generator
        By default numpy.random.
    """

    def __init__(self, blmin, n_top, rattle_strength=0.8,
                 rattle_prop=0.4, test_dist_to_slab=True, use_tags=False,
                 verbose=False, rng=np.random):
        OffspringCreator.__init__(self, verbose, rng=rng)
        self.blmin = blmin
        self.n_top = n_top
        self.rattle_strength = rattle_strength
        self.rattle_prop = rattle_prop
        self.test_dist_to_slab = test_dist_to_slab
        self.use_tags = use_tags

        self.descriptor = 'RattleMutation'
        self.min_inputs = 1

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.mutate(f)
        if indi is None:
            return indi, 'mutation: rattle'

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [f.info['confid']]

        return self.finalize_individual(indi), 'mutation: rattle'

    def mutate(self, atoms):
        """Does the actual mutation."""
        N = len(atoms) if self.n_top is None else self.n_top
        slab = atoms[:len(atoms) - N]
        atoms = atoms[-N:]
        tags = atoms.get_tags() if self.use_tags else np.arange(N)
        pos_ref = atoms.get_positions()
        num = atoms.get_atomic_numbers()
        cell = atoms.get_cell()
        pbc = atoms.get_pbc()
        st = 2. * self.rattle_strength

        count = 0
        maxcount = 1000
        too_close = True
        while too_close and count < maxcount:
            count += 1
            pos = pos_ref.copy()
            ok = False
            for tag in np.unique(tags):
                select = np.where(tags == tag)
                if self.rng.random() < self.rattle_prop:
                    ok = True
                    r = self.rng.random(3)
                    pos[select] += st * (r - 0.5)

            if not ok:
                # Nothing got rattled
                continue

            top = Atoms(num, positions=pos, cell=cell, pbc=pbc, tags=tags)
            too_close = atoms_too_close(
                top, self.blmin, use_tags=self.use_tags)
            if not too_close and self.test_dist_to_slab:
                too_close = atoms_too_close_two_sets(top, slab, self.blmin)

        if count == maxcount:
            return None

        mutant = slab + top
        return mutant


class PermutationMutation(OffspringCreator):
    """Mutation that permutes a percentage of the atom types in the cluster.

    Parameters:

    n_top: Number of atoms optimized by the GA.

    probability: The probability with which an atom is permuted.

    test_dist_to_slab: whether to also make sure that the distances
        between the atoms and the slab satisfy the blmin.

    use_tags: if True, the atomic tags will be used to preserve
        molecular identity. Permutations will then happen
        at the molecular level, i.e. swapping the center-of-
        positions of two moieties while preserving their
        internal geometries.

    blmin: Dictionary defining the minimum distance between atoms
        after the permutation. If equal to None (the default),
        no such check is performed.

    rng: Random number generator
        By default numpy.random.
    """

    def __init__(self, n_top, probability=0.33, test_dist_to_slab=True,
                 use_tags=False, blmin=None, rng=np.random, verbose=False):
        OffspringCreator.__init__(self, verbose, rng=rng)
        self.n_top = n_top
        self.probability = probability
        self.test_dist_to_slab = test_dist_to_slab
        self.use_tags = use_tags
        self.blmin = blmin

        self.descriptor = 'PermutationMutation'
        self.min_inputs = 1

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.mutate(f)
        if indi is None:
            return indi, 'mutation: permutation'

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [f.info['confid']]

        return self.finalize_individual(indi), 'mutation: permutation'

    def mutate(self, atoms):
        """Does the actual mutation."""
        N = len(atoms) if self.n_top is None else self.n_top
        slab = atoms[:len(atoms) - N]
        atoms = atoms[-N:]
        if self.use_tags:
            gather_atoms_by_tag(atoms)
        tags = atoms.get_tags() if self.use_tags else np.arange(N)
        pos_ref = atoms.get_positions()
        num = atoms.get_atomic_numbers()
        cell = atoms.get_cell()
        pbc = atoms.get_pbc()
        symbols = atoms.get_chemical_symbols()

        unique_tags = np.unique(tags)
        n = len(unique_tags)
        swaps = int(np.ceil(n * self.probability / 2.))

        sym = []
        for tag in unique_tags:
            indices = np.where(tags == tag)[0]
            s = ''.join([symbols[j] for j in indices])
            sym.append(s)

        assert len(np.unique(sym)) > 1, \
            'Permutations with one atom (or molecule) type is not valid'

        count = 0
        maxcount = 1000
        too_close = True
        while too_close and count < maxcount:
            count += 1
            pos = pos_ref.copy()
            for _ in range(swaps):
                i = j = 0
                while sym[i] == sym[j]:
                    i = self.rng.randint(0, high=n)
                    j = self.rng.randint(0, high=n)
                ind1 = np.where(tags == i)
                ind2 = np.where(tags == j)
                cop1 = np.mean(pos[ind1], axis=0)
                cop2 = np.mean(pos[ind2], axis=0)
                pos[ind1] += cop2 - cop1
                pos[ind2] += cop1 - cop2

            top = Atoms(num, positions=pos, cell=cell, pbc=pbc, tags=tags)
            if self.blmin is None:
                too_close = False
            else:
                too_close = atoms_too_close(
                    top, self.blmin, use_tags=self.use_tags)
                if not too_close and self.test_dist_to_slab:
                    too_close = atoms_too_close_two_sets(top, slab, self.blmin)

        if count == maxcount:
            return None

        mutant = slab + top
        return mutant


class MirrorMutation(OffspringCreator):
    """A mirror mutation, as described in
    TO BE PUBLISHED.

    This mutation mirrors half of the cluster in a
    randomly oriented cutting plane discarding the other half.

    Parameters:

    blmin: Dictionary defining the minimum allowed
        distance between atoms.

    n_top: Number of atoms the GA optimizes.

    reflect: Defines if the mirrored half is also reflected
        perpendicular to the mirroring plane.

    rng: Random number generator
        By default numpy.random.
    """

    def __init__(self, blmin, n_top, reflect=False, rng=np.random,
                 verbose=False):
        OffspringCreator.__init__(self, verbose, rng=rng)
        self.blmin = blmin
        self.n_top = n_top
        self.reflect = reflect

        self.descriptor = 'MirrorMutation'
        self.min_inputs = 1

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.mutate(f)
        if indi is None:
            return indi, 'mutation: mirror'

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [f.info['confid']]

        return self.finalize_individual(indi), 'mutation: mirror'

    def mutate(self, atoms):
        """ Do the mutation of the atoms input. """
        reflect = self.reflect
        tc = True
        slab = atoms[0:len(atoms) - self.n_top]
        top = atoms[len(atoms) - self.n_top: len(atoms)]
        num = top.numbers
        unique_types = list(set(num))
        nu = {u: sum(num == u) for u in unique_types}
        n_tries = 1000
        counter = 0
        changed = False

        while tc and counter < n_tries:
            counter += 1
            cand = top.copy()
            pos = cand.get_positions()

            cm = np.average(top.get_positions(), axis=0)

            # first select a randomly oriented cutting plane
            theta = pi * self.rng.random()
            phi = 2. * pi * self.rng.random()
            n = (cos(phi) * sin(theta), sin(phi) * sin(theta), cos(theta))
            n = np.array(n)

            # Calculate all atoms signed distance to the cutting plane
            D = []
            for (i, p) in enumerate(pos):
                d = np.dot(p - cm, n)
                D.append((i, d))

            # Sort the atoms by their signed distance
            D.sort(key=lambda x: x[1])
            nu_taken = {}

            # Select half of the atoms needed for a full cluster
            p_use = []
            n_use = []
            for (i, d) in D:
                if num[i] not in nu_taken.keys():
                    nu_taken[num[i]] = 0
                if nu_taken[num[i]] < nu[num[i]] / 2.:
                    p_use.append(pos[i])
                    n_use.append(num[i])
                    nu_taken[num[i]] += 1

            # calculate the mirrored position and add these.
            pn = []
            for p in p_use:
                pt = p - 2. * np.dot(p - cm, n) * n
                if reflect:
                    pt = -pt + 2 * cm + 2 * n * np.dot(pt - cm, n)
                pn.append(pt)

            n_use.extend(n_use)
            p_use.extend(pn)

            # In the case of an uneven number of
            # atoms we need to add one extra
            for n in nu:
                if nu[n] % 2 == 0:
                    continue
                while sum(n_use == n) > nu[n]:
                    for i in range(int(len(n_use) / 2), len(n_use)):
                        if n_use[i] == n:
                            del p_use[i]
                            del n_use[i]
                            break
                assert sum(n_use == n) == nu[n]

            # Make sure we have the correct number of atoms
            # and rearrange the atoms so they are in the right order
            for i in range(len(n_use)):
                if num[i] == n_use[i]:
                    continue
                for j in range(i + 1, len(n_use)):
                    if n_use[j] == num[i]:
                        tn = n_use[i]
                        tp = p_use[i]
                        n_use[i] = n_use[j]
                        p_use[i] = p_use[j]
                        p_use[j] = tp
                        n_use[j] = tn

            # Finally we check that nothing is too close in the end product.
            cand = Atoms(num, p_use, cell=slab.get_cell(), pbc=slab.get_pbc())

            tc = atoms_too_close(cand, self.blmin)
            if tc:
                continue
            tc = atoms_too_close_two_sets(slab, cand, self.blmin)

            if not changed and counter > n_tries // 2:
                reflect = not reflect
                changed = True

            tot = slab + cand

        if counter == n_tries:
            return None
        return tot


class StrainMutation(OffspringCreator):
    """ Mutates a candidate by applying a randomly generated strain.

    For more information, see also:

      * :doi:`Glass, Oganov, Hansen, Comp. Phys. Comm. 175 (2006) 713-720
        <10.1016/j.cpc.2006.07.020>`

      * :doi:`Lonie, Zurek, Comp. Phys. Comm. 182 (2011) 372-387
        <10.1016/j.cpc.2010.07.048>`

    After initialization of the mutation, a scaling volume
    (to which each mutated structure is scaled before checking the
    constraints) is typically generated from the population,
    which is then also occasionally updated in the course of the
    GA run.

    Parameters:

    blmin: dict
        The closest allowed interatomic distances on the form:
        {(Z, Z*): dist, ...}, where Z and Z* are atomic numbers.

    cellbounds: ase.ga.utilities.CellBounds instance
        Describes limits on the cell shape, see
        :class:`~ase.ga.utilities.CellBounds`.

    stddev: float
        Standard deviation used in the generation of the
        strain matrix elements.

    number_of_variable_cell_vectors: int (default 3)
        The number of variable cell vectors (1, 2 or 3).
        To keep things simple, it is the 'first' vectors which
        will be treated as variable, i.e. the 'a' vector in the
        univariate case, the 'a' and 'b' vectors in the bivariate
        case, etc.

    use_tags: boolean
        Whether to use the atomic tags to preserve molecular identity.

    rng: Random number generator
        By default numpy.random.
    """

    def __init__(self, blmin, cellbounds=None, stddev=0.7,
                 number_of_variable_cell_vectors=3, use_tags=False,
                 rng=np.random, verbose=False):
        OffspringCreator.__init__(self, verbose, rng=rng)
        self.blmin = blmin
        self.cellbounds = cellbounds
        self.stddev = stddev
        self.number_of_variable_cell_vectors = number_of_variable_cell_vectors
        self.use_tags = use_tags

        self.scaling_volume = None
        self.descriptor = 'StrainMutation'
        self.min_inputs = 1

    def update_scaling_volume(self, population, w_adapt=0.5, n_adapt=0):
        """Function to initialize or update the scaling volume in a GA run.

        w_adapt: weight of the new vs the old scaling volume

        n_adapt: number of best candidates in the population that
                 are used to calculate the new scaling volume
        """
        if not n_adapt:
            # if not set, take best 20% of the population
            n_adapt = int(np.ceil(0.2 * len(population)))
        v_new = np.mean([a.get_volume() for a in population[:n_adapt]])

        if not self.scaling_volume:
            self.scaling_volume = v_new
        else:
            volumes = [self.scaling_volume, v_new]
            weights = [1 - w_adapt, w_adapt]
            self.scaling_volume = np.average(volumes, weights=weights)

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.mutate(f)
        if indi is None:
            return indi, 'mutation: strain'

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [f.info['confid']]

        return self.finalize_individual(indi), 'mutation: strain'

    def mutate(self, atoms):
        """ Does the actual mutation. """
        cell_ref = atoms.get_cell()
        pos_ref = atoms.get_positions()

        if self.scaling_volume is None:
            # The scaling_volume has not been set (yet),
            # so we give it the same volume as the parent
            vol_ref = atoms.get_volume()
        else:
            vol_ref = self.scaling_volume

        if self.use_tags:
            tags = atoms.get_tags()
            gather_atoms_by_tag(atoms)
            pos = atoms.get_positions()

        mutant = atoms.copy()

        count = 0
        too_close = True
        maxcount = 1000
        while too_close and count < maxcount:
            count += 1

            # generating the strain matrix:
            strain = np.identity(3)
            for i in range(self.number_of_variable_cell_vectors):
                for j in range(i + 1):
                    r = self.rng.normal(loc=0., scale=self.stddev)
                    if i == j:
                        strain[i, j] += r
                    else:
                        epsilon = 0.5 * r
                        strain[i, j] += epsilon
                        strain[j, i] += epsilon

            # applying the strain:
            cell_new = np.dot(strain, cell_ref)

            # convert the submatrix with the variable cell vectors
            # to a lower triangular form
            cell_new = calc_rotated_cell(cell_new)
            for i in range(self.number_of_variable_cell_vectors, 3):
                cell_new[i] = cell_ref[i]

            cell_new = Cell(cell_new)

            # volume scaling:
            if self.number_of_variable_cell_vectors > 0:
                scaling = vol_ref / cell_new.volume
                scaling **= 1. / self.number_of_variable_cell_vectors
                cell_new[:self.number_of_variable_cell_vectors] *= scaling

            # check cell dimensions:
            if not self.cellbounds.is_within_bounds(cell_new):
                continue

            # ensure non-variable cell vectors are indeed unchanged
            for i in range(self.number_of_variable_cell_vectors, 3):
                assert np.allclose(cell_new[i], cell_ref[i])

            # check that the volume is correct
            assert np.isclose(vol_ref, cell_new.volume)

            # apply the new unit cell and scale
            # the atomic positions accordingly
            mutant.set_cell(cell_ref, scale_atoms=False)

            if self.use_tags:
                transfo = np.linalg.solve(cell_ref, cell_new)
                for tag in np.unique(tags):
                    select = np.where(tags == tag)
                    cop = np.mean(pos[select], axis=0)
                    disp = np.dot(cop, transfo) - cop
                    mutant.positions[select] += disp
            else:
                mutant.set_positions(pos_ref)

            mutant.set_cell(cell_new, scale_atoms=not self.use_tags)
            mutant.wrap()

            # check the interatomic distances
            too_close = atoms_too_close(mutant, self.blmin,
                                        use_tags=self.use_tags)

        if count == maxcount:
            mutant = None

        return mutant


class PermuStrainMutation(CombinationMutation):
    """Combination of PermutationMutation and StrainMutation.

    For more information, see also:

      * :doi:`Lonie, Zurek, Comp. Phys. Comm. 182 (2011) 372-387
        <10.1016/j.cpc.2010.07.048>`

    Parameters:

    permutationmutation: OffspringCreator instance
        A mutation that permutes atom types.

    strainmutation: OffspringCreator instance
        A mutation that mutates by straining.
    """

    def __init__(self, permutationmutation, strainmutation, verbose=False):
        super().__init__(permutationmutation,
                         strainmutation,
                         verbose=verbose)
        self.descriptor = 'permustrain'


class RotationalMutation(OffspringCreator):
    """Mutates a candidate by applying random rotations
    to multi-atom moieties in the structure (atoms with
    the same tag are considered part of one such moiety).

    Only performs whole-molecule rotations, no internal
    rotations.

    For more information, see also:

      * `Zhu Q., Oganov A.R., Glass C.W., Stokes H.T,
        Acta Cryst. (2012), B68, 215-226.`__

        __ https://dx.doi.org/10.1107/S0108768112017466

    Parameters:

    blmin: dict
        The closest allowed interatomic distances on the form:
        {(Z, Z*): dist, ...}, where Z and Z* are atomic numbers.

    n_top: int or None
        The number of atoms to optimize (None = include all).

    fraction: float
        Fraction of the moieties to be rotated.

    tags: None or list of integers
        Specifies, respectively, whether all moieties or only those
        with matching tags are eligible for rotation.

    min_angle: float
        Minimal angle (in radians) for each rotation;
        should lie in the interval [0, pi].

    test_dist_to_slab: boolean
        Whether also the distances to the slab
        should be checked to satisfy the blmin.

    rng: Random number generator
        By default numpy.random.
    """

    def __init__(self, blmin, n_top=None, fraction=0.33, tags=None,
                 min_angle=1.57, test_dist_to_slab=True, rng=np.random,
                 verbose=False):
        OffspringCreator.__init__(self, verbose, rng=rng)
        self.blmin = blmin
        self.n_top = n_top
        self.fraction = fraction
        self.tags = tags
        self.min_angle = min_angle
        self.test_dist_to_slab = test_dist_to_slab
        self.descriptor = 'RotationalMutation'
        self.min_inputs = 1

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.mutate(f)
        if indi is None:
            return indi, 'mutation: rotational'

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [f.info['confid']]

        return self.finalize_individual(indi), 'mutation: rotational'

    def mutate(self, atoms):
        """Does the actual mutation."""
        N = len(atoms) if self.n_top is None else self.n_top
        slab = atoms[:len(atoms) - N]
        atoms = atoms[-N:]

        mutant = atoms.copy()
        gather_atoms_by_tag(mutant)
        pos = mutant.get_positions()
        tags = mutant.get_tags()
        eligible_tags = tags if self.tags is None else self.tags

        indices = {}
        for tag in np.unique(tags):
            hits = np.where(tags == tag)[0]
            if len(hits) > 1 and tag in eligible_tags:
                indices[tag] = hits

        n_rot = int(np.ceil(len(indices) * self.fraction))
        chosen_tags = self.rng.choice(list(indices.keys()), size=n_rot,
                                      replace=False)

        too_close = True
        count = 0
        maxcount = 10000
        while too_close and count < maxcount:
            newpos = np.copy(pos)
            for tag in chosen_tags:
                p = np.copy(newpos[indices[tag]])
                cop = np.mean(p, axis=0)

                if len(p) == 2:
                    line = (p[1] - p[0]) / np.linalg.norm(p[1] - p[0])
                    while True:
                        axis = self.rng.random(3)
                        axis /= np.linalg.norm(axis)
                        a = np.arccos(np.dot(axis, line))
                        if np.pi / 4 < a < np.pi * 3 / 4:
                            break
                else:
                    axis = self.rng.random(3)
                    axis /= np.linalg.norm(axis)

                angle = self.min_angle
                angle += 2 * (np.pi - self.min_angle) * self.rng.random()

                m = get_rotation_matrix(axis, angle)
                newpos[indices[tag]] = np.dot(m, (p - cop).T).T + cop

            mutant.set_positions(newpos)
            mutant.wrap()
            too_close = atoms_too_close(mutant, self.blmin, use_tags=True)
            count += 1

            if not too_close and self.test_dist_to_slab:
                too_close = atoms_too_close_two_sets(slab, mutant, self.blmin)

        if count == maxcount:
            mutant = None
        else:
            mutant = slab + mutant

        return mutant


class RattleRotationalMutation(CombinationMutation):
    """Combination of RattleMutation and RotationalMutation.

    Parameters:

    rattlemutation: OffspringCreator instance
        A mutation that rattles atoms.

    rotationalmutation: OffspringCreator instance
        A mutation that rotates moieties.
    """

    def __init__(self, rattlemutation, rotationalmutation, verbose=False):
        super().__init__(rattlemutation,
                         rotationalmutation,
                         verbose=verbose)
        self.descriptor = 'rattlerotational'



from os.path import exists

import numpy as np

from ase.calculators.calculator import Calculator, all_changes
from ase.io.trajectory import Trajectory
from ase.parallel import broadcast, world
from ase.units import fs, kJ, mol, nm


def restart_from_trajectory(prev_traj, *args, prev_steps=None, atoms=None,
                            **kwargs):
    """This function helps the user to restart a plumed simulation
    from a trajectory file.

    Parameters
    ----------
    prev_traj : Trajectory object
        previous simulated trajectory

    prev_steps : int. Default steps in prev_traj.
        number of previous steps

    others :
       Same parameters of :mod:`~ase.calculators.plumed` calculator

    Returns
    -------
    Plumed calculator


    .. note:: prev_steps is crucial when trajectory does not contain
        all the previous steps.
    """
    atoms.calc = Plumed(*args, atoms=atoms, restart=True, **kwargs)

    with Trajectory(prev_traj) as traj:
        if prev_steps is None:
            atoms.calc.istep = len(traj) - 1
        else:
            atoms.calc.istep = prev_steps
        atoms.set_positions(traj[-1].get_positions())
        atoms.set_momenta(traj[-1].get_momenta())
    return atoms.calc


class Plumed(Calculator):
    implemented_properties = ['energy', 'forces']

    def __init__(self, calc, input, timestep, atoms=None, kT=1., log='',
                 restart=False, use_charge=False, update_charge=False):
        """
        Plumed calculator is used for simulations of enhanced sampling methods
        with the open-source code PLUMED (plumed.org).

        [1] The PLUMED consortium, Nat. Methods 16, 670 (2019)
        [2] Tribello, Bonomi, Branduardi, Camilloni, and Bussi,
        Comput. Phys. Commun. 185, 604 (2014)

        Parameters
        ----------
        calc: Calculator object
            It  computes the unbiased forces

        input: List of strings
            It contains the setup of plumed actions

        timestep: float
            Time step of the simulated dynamics

        atoms: Atoms
            Atoms object to be attached

        kT: float. Default 1.
            Value of the thermal energy in eV units. It is important for
            some methods of plumed like Well-Tempered Metadynamics.

        log: string
            Log file of the plumed calculations

        restart: boolean. Default False
            True if the simulation is restarted.

        use_charge: boolean. Default False
            True if you use some collective variable which needs charges. If
            use_charges is True and update_charge is False, you have to define
            initial charges and then this charge will be used during all
            simulation.

        update_charge: boolean. Default False
            True if you want the charges to be updated each time step. This
            will fail in case that calc does not have 'charges' in its
            properties.


        .. note:: For this case, the calculator is defined strictly with the
            object atoms inside. This is necessary for initializing the
            Plumed object. For conserving ASE convention, it can be initialized
            as atoms.calc = (..., atoms=atoms, ...)


        .. note:: In order to guarantee a proper restart, the user has to fix
            momenta, positions and Plumed.istep, where the positions and
            momenta corresponds to the last configuration in the previous
            simulation, while Plumed.istep is the number of timesteps
            performed previously. This can be done using
            ase.calculators.plumed.restart_from_trajectory.
        """

        from plumed import Plumed as pl

        if atoms is None:
            raise TypeError('plumed calculator has to be defined with the \
                             object atoms inside.')

        self.istep = 0
        Calculator.__init__(self, atoms=atoms)

        self.input = input
        self.calc = calc
        self.use_charge = use_charge
        self.update_charge = update_charge

        if world.rank == 0:
            natoms = len(atoms.get_positions())
            self.plumed = pl()

            ''' Units setup
            warning: inputs and outputs of plumed will still be in
            plumed units.

            The change of Plumed units to ASE units is:
            kjoule/mol to eV
            nm to Angstrom
            ps to ASE time units
            ASE and plumed - charge unit is in e units
            ASE and plumed - mass unit is in a.m.u units '''

            ps = 1000 * fs
            self.plumed.cmd("setMDEnergyUnits", mol / kJ)
            self.plumed.cmd("setMDLengthUnits", 1 / nm)
            self.plumed.cmd("setMDTimeUnits", 1 / ps)
            self.plumed.cmd("setMDChargeUnits", 1.)
            self.plumed.cmd("setMDMassUnits", 1.)

            self.plumed.cmd("setNatoms", natoms)
            self.plumed.cmd("setMDEngine", "ASE")
            self.plumed.cmd("setLogFile", log)
            self.plumed.cmd("setTimestep", float(timestep))
            self.plumed.cmd("setRestart", restart)
            self.plumed.cmd("setKbT", float(kT))
            self.plumed.cmd("init")
            for line in input:
                self.plumed.cmd("readInputLine", line)
        self.atoms = atoms

    def _get_name(self):
        return f'{self.calc.name}+Plumed'

    def calculate(self, atoms=None, properties=['energy', 'forces'],
                  system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        comp = self.compute_energy_and_forces(self.atoms.get_positions(),
                                              self.istep)
        energy, forces = comp
        self.istep += 1
        self.results['energy'], self. results['forces'] = energy, forces

    def compute_energy_and_forces(self, pos, istep):
        unbiased_energy = self.calc.get_potential_energy(self.atoms)
        unbiased_forces = self.calc.get_forces(self.atoms)

        if world.rank == 0:
            ener_forc = self.compute_bias(pos, istep, unbiased_energy)
        else:
            ener_forc = None
        energy_bias, forces_bias = broadcast(ener_forc)
        energy = unbiased_energy + energy_bias
        forces = unbiased_forces + forces_bias
        return energy, forces

    def compute_bias(self, pos, istep, unbiased_energy):
        self.plumed.cmd("setStep", istep)

        if self.use_charge:
            if 'charges' in self.calc.implemented_properties and \
               self.update_charge:
                charges = self.calc.get_charges(atoms=self.atoms.copy())

            elif self.atoms.has('initial_charges') and not self.update_charge:
                charges = self.atoms.get_initial_charges()

            else:
                assert not self.update_charge, "Charges cannot be updated"
                assert self.update_charge, "Not initial charges in Atoms"

            self.plumed.cmd("setCharges", charges)

        # Box for functions with PBC in plumed
        if self.atoms.cell:
            cell = np.asarray(self.atoms.get_cell())
            self.plumed.cmd("setBox", cell)

        self.plumed.cmd("setPositions", pos)
        self.plumed.cmd("setEnergy", unbiased_energy)
        self.plumed.cmd("setMasses", self.atoms.get_masses())
        forces_bias = np.zeros((self.atoms.get_positions()).shape)
        self.plumed.cmd("setForces", forces_bias)
        virial = np.zeros((3, 3))
        self.plumed.cmd("setVirial", virial)
        self.plumed.cmd("prepareCalc")
        self.plumed.cmd("performCalc")
        energy_bias = np.zeros((1,))
        self.plumed.cmd("getBias", energy_bias)
        return [energy_bias, forces_bias]

    def write_plumed_files(self, images):
        """ This function computes what is required in
        plumed input for some trajectory.

        The outputs are saved in the typical files of
        plumed such as COLVAR, HILLS """
        for i, image in enumerate(images):
            pos = image.get_positions()
            self.compute_energy_and_forces(pos, i)
        return self.read_plumed_files()

    def read_plumed_files(self, file_name=None):
        read_files = {}
        if file_name is not None:
            read_files[file_name] = np.loadtxt(file_name, unpack=True)
        else:
            for line in self.input:
                if line.find('FILE') != -1:
                    ini = line.find('FILE')
                    end = line.find(' ', ini)
                    if end == -1:
                        file_name = line[ini + 5:]
                    else:
                        file_name = line[ini + 5:end]
                    read_files[file_name] = np.loadtxt(file_name, unpack=True)

            if len(read_files) == 0:
                if exists('COLVAR'):
                    read_files['COLVAR'] = np.loadtxt('COLVAR', unpack=True)
                if exists('HILLS'):
                    read_files['HILLS'] = np.loadtxt('HILLS', unpack=True)
        assert len(read_files) != 0, "There are not files for reading"
        return read_files

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.plumed.finalize()



import warnings

from ase.calculators.calculator import FileIOCalculator
from ase.io import read, write
from ase.io.gamess_us import clean_userscr, get_userscr


class GAMESSUS(FileIOCalculator):
    implemented_properties = ['energy', 'forces', 'dipole']
    _legacy_default_command = 'rungms PREFIX.inp > PREFIX.log 2> PREFIX.err'
    discard_results_on_any_change = True

    def __init__(self, restart=None,
                 ignore_bad_restart_file=FileIOCalculator._deprecated,
                 label='gamess_us', atoms=None, command=None, userscr=None,
                 **kwargs):
        """
        GAMESS-US keywords are specified using dictionaries of keywords.
        For example, to run a CCSD(T)/cc-pVDZ calculation, you might use the
        following arguments:

            >>> calc = GAMESSUS(contrl={'scftyp': 'rhf', 'cctyp': 'ccsd(t)',
            >>>                         'ispher': 1, 'runtyp': 'energy'},
            >>>                 basis={'gbasis': 'CCD'})

        This will create an input file that looks like the following:

            >>>  $CONTRL
            >>>   SCFTYP=RHF
            >>>   CCTYP=CCSD(T)
            >>>   ISPHER=1
            >>>   RUNTYP=ENERGY
            >>>  $END
            >>>
            >>>  $BASIS
            >>>   GBASIS=CCSD
            >>>  $END

        See the INPUT.DOC file provided by GAMESS-US for more information.

        If `runtyp` is not specified, it will be set automatically.

        If no basis is specified, 3-21G will be used by default.
        A dictionary containing literal per-index or per-element basis sets
        can be passed to the `basis` keyword. This will result in the basis
        set being printed in the $DATA block, alongside the geometry
        specification.
        Otherwise, `basis` is assumed to contain keywords for the $BASIS
        block, such as GBASIS and NGAUSS.

        If a multiplicity is not set in contrl['mult'], the multiplicity
        will be guessed based on the Atoms object's initial magnetic moments.

        The GAMESSUS calculator has some special keyword:

        xc: str
            The exchange-correlation functional to use for DFT calculations.
            In most circumstances, setting xc is equivalent to setting
            contrl['dfttyp']. xc will be ignored if a value has also
            been provided to contrl['dfttyp'].

        userscr: str
            The location of the USERSCR directory specified in your
            `rungms` script. If not provided, an attempt will be made
            to automatically determine the location of this directory.
            If this fails, you will need to manually specify the path
            to this directory to the calculator using this keyword.

        ecp: dict
            A per-index or per-element dictionary of ECP specifications.
        """

        # Backward compatibility fix:
        if command is None and 'ASE_GAMESSUS_COMMAND' in self.cfg:
            command = self.cfg['ASE_GAMESSUS_COMMAND']

        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file,
                                  label, atoms, command, **kwargs)
        self.userscr = userscr

    def _get_name(self):
        return 'gamess_us'

    def calculate(self, *args, **kwargs):
        if self.userscr is None:
            if 'rungms' in self.command:
                self.userscr = get_userscr(self.prefix, self.command)

        if self.userscr is None:
            warnings.warn("Could not determine USERSCR! "
                          "GAMESS may refuse to run more than once for "
                          "this job. Please pass userscr to the GAMESSUS "
                          "Calculator if you run into problems!")
        else:
            clean_userscr(self.userscr, self.prefix)

        FileIOCalculator.calculate(self, *args, **kwargs)

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        write(self.label + '.inp', atoms, properties=properties,
              format='gamess-us-in', **self.parameters)

    def read_results(self):
        output = read(self.label + '.log')
        self.calc = output.calc
        self.results = output.calc.results



import pytest

from ase import io
from ase.build import molecule


@pytest.mark.calculator('gpaw')
@pytest.mark.filterwarnings('ignore:The keyword')
@pytest.mark.filterwarnings('ignore:convert_string_to_fd')
# Ignore calculator constructor keyword warning for now
def test_no_spin_and_spin(factory):
    txt = 'out.txt'

    calculator = factory.calc(mode='fd', h=0.3, txt=txt)
    atoms = molecule('H2', calculator=calculator)
    atoms.center(vacuum=3)
    atoms.get_potential_energy()
    atoms.set_initial_magnetic_moments([0.5, 0.5])
    calculator.set(charge=1)
    atoms.get_potential_energy()

    # read again
    t = io.read(txt, index=':')
    assert isinstance(t, list)
    M = t[1].get_magnetic_moments()
    assert abs(M - 0.2).max() < 0.1



from typing import Iterable

import numpy as np
import pytest

from ase.spectrum.doscollection import (
    DOSCollection,
    GridDOSCollection,
    RawDOSCollection,
)
from ase.spectrum.dosdata import DOSData, GridDOSData, RawDOSData


class MinimalDOSCollection(DOSCollection):
    """Inherit from abstract base class to check its features"""

    def __init__(self, dos_series: Iterable[DOSData]) -> None:
        super().__init__(dos_series)


class YetAnotherDOSCollection(DOSCollection):
    """Inherit from abstract base class to check its features"""

    def __init__(self, dos_series: Iterable[DOSData]) -> None:
        super().__init__(dos_series)


class TestDOSCollection:
    @pytest.fixture()
    def rawdos(self):
        return RawDOSData([1., 2., 4.], [2., 3., 2.],
                          info={'my_key': 'my_value'})

    @pytest.fixture()
    def another_rawdos(self):
        return RawDOSData([3., 2., 5.], [1., 0., 2.],
                          info={'other_key': 'other_value'})

    @pytest.fixture()
    def mindoscollection(self, rawdos, another_rawdos):
        return MinimalDOSCollection([rawdos, another_rawdos])

    @pytest.mark.parametrize('n_entries', [0, 1, 3])
    def test_sequence(self, rawdos, n_entries):
        dc = MinimalDOSCollection([rawdos] * n_entries)
        assert len(dc) == n_entries
        for i in range(n_entries):
            assert dc[i] == rawdos

        with pytest.raises(IndexError):
            dc[n_entries + 1]
        with pytest.raises(TypeError):
            dc['hello']

    linewidths = [1, 5, None]

    @pytest.mark.usefixtures("figure")
    @pytest.mark.parametrize('linewidth', linewidths)
    def test_plot(self, mindoscollection, figure, linewidth):
        npts = 20

        if linewidth is None:
            mplargs = None
        else:
            mplargs = {'linewidth': linewidth}

        ax = figure.add_subplot(111)

        with pytest.warns(UserWarning):  # Default width is small for npts=20
            ax_out = mindoscollection.plot(npts=npts, ax=ax, mplargs=mplargs)
        assert ax_out == ax

        assert ([line.get_label() for line in ax.get_legend().get_lines()]
                == ['my_key: my_value', 'other_key: other_value'])

    def test_slicing(self, rawdos, another_rawdos):
        dc = MinimalDOSCollection([rawdos, another_rawdos, rawdos])
        assert dc[1:]._almost_equals(
            MinimalDOSCollection([another_rawdos, rawdos]))
        assert dc[:-1]._almost_equals(
            MinimalDOSCollection([rawdos, another_rawdos]))

    # It would be much nicer if this test could be done with parameterization,
    # but creating equality_data as a parameter list requires the lazy_fixtures
    # pytest plugin.
    def test_collection_equality(self, rawdos, another_rawdos):
        equality_data = [([], [], True),
                         ([rawdos], [rawdos], True),
                         ([rawdos, another_rawdos],
                          [rawdos, another_rawdos], True),
                         ([], [rawdos], False),
                         ([rawdos], [], False),
                         ([rawdos, another_rawdos], [rawdos], False),
                         ([rawdos, another_rawdos],
                         [another_rawdos, rawdos], False)]

        for series_1, series_2, isequal in equality_data:
            assert (MinimalDOSCollection(series_1)
                    ._almost_equals(MinimalDOSCollection(series_2)) == isequal)

    @pytest.mark.parametrize('other', [True, 1, 0.5, 'string', rawdos])
    def test_equality_wrongtype(self, rawdos, other):
        assert not MinimalDOSCollection([rawdos])._almost_equals(other)

    def test_addition(self, rawdos, another_rawdos):
        dc = MinimalDOSCollection([rawdos])

        double_dc = dc + dc
        assert len(double_dc) == 2
        assert double_dc[0]._almost_equals(rawdos)
        assert double_dc[1]._almost_equals(rawdos)

        assert (dc + MinimalDOSCollection([another_rawdos])
                )._almost_equals(dc + another_rawdos)

        with pytest.raises(TypeError):
            MinimalDOSCollection([rawdos]) + YetAnotherDOSCollection([rawdos])
        with pytest.raises(TypeError):
            MinimalDOSCollection([rawdos]) + 'string'

    @pytest.mark.parametrize('options', [{'energies': [1., 1.1, 1.2],
                                          'width': 1.3,
                                          'smearing': 'Gauss'},
                                         {'energies': [1.7, 2.1, 2.0],
                                          'width': 3.4,
                                          'smearing': 'Gauss'}])
    def test_sample(self, rawdos, another_rawdos, options):
        dc = MinimalDOSCollection([rawdos, another_rawdos])
        sampled_data = dc._sample(**options)
        for i, data in enumerate((rawdos, another_rawdos)):
            # Check consistency with individual DOSData objects
            newdos_weights = data._sample(**options)
            assert np.allclose(sampled_data[i, :], newdos_weights)
            # Check we aren't trivially comparing zeros
            assert np.all(sampled_data)

    sample_grid_options = [{'npts': 10, 'xmin': -2, 'xmax': 10,
                            'padding': 3, 'width': 1},
                           {'npts': 12, 'xmin': 0, 'xmax': 4,
                            'padding': 2.1, 'width': 2.3}]

    @pytest.mark.parametrize('options', sample_grid_options)
    def test_sample_grid(self, rawdos, another_rawdos, options):
        ref_min = min(rawdos.get_energies())
        ref_max = max(another_rawdos.get_energies())

        # Check auto minimum
        dc = MinimalDOSCollection([rawdos, another_rawdos])
        dos = dc.sample_grid(10, xmax=options['xmax'],
                             padding=options['padding'],
                             width=options['width'])
        energies = dos.get_energies()

        assert (pytest.approx(energies[0])
                == ref_min - options['padding'] * options['width'])
        assert pytest.approx(energies[-1]) == options['xmax']

        # Check auto maximum
        dos = dc.sample_grid(10, xmin=options['xmin'],
                             padding=options['padding'],
                             width=options['width'])
        energies = dos.get_energies()
        assert pytest.approx(energies[0]) == options['xmin']
        assert (pytest.approx(energies[-1])
                == ref_max + options['padding'] * options['width'])

        # Check values
        dos = dc.sample_grid(**options)
        energies = dos.get_energies()
        weights = dos.get_all_weights()
        for i, data in enumerate((rawdos, another_rawdos)):
            tmp_dos = data.sample_grid(**options)
            tmp_weights = tmp_dos.get_weights()
            assert np.allclose(weights[i, :], tmp_weights)

    def test_sample_empty(self):
        empty_dc = MinimalDOSCollection([])
        with pytest.raises(IndexError):
            empty_dc._sample(10)
        with pytest.raises(IndexError):
            empty_dc.sample_grid(10)

    @pytest.mark.parametrize('x, weights, bad_info',
                             [([1, 2, 4, 5],
                               [[0, 1, 1, 0], [2, 1, 2, 1]],
                               [{'notenough': 'entries'}]),
                              ([3.1, 2.4, 1.1],
                               [[2, 1., 3.12]],
                               [{'too': 'many'}, {'entries': 'here'}])
                              ])
    def test_from_data(self, x, weights, bad_info):
        dc = DOSCollection.from_data(x, weights)

        for i, dos_data in enumerate(dc):
            assert dos_data.info == {}
            assert np.allclose(dos_data.get_energies(), x)
            assert np.allclose(dos_data.get_weights(), weights[i])

        with pytest.raises(ValueError):
            dc = DOSCollection.from_data(x, weights, info=bad_info)

    collection_data = [[([1., 2., 3.], [1., 1., 2.])],
                       [([1., 2., 3.], [1., 1., 2.]),
                        ([2., 3.5], [0.5, 1.])],
                       [([1., 2., 3.], [1., 1., 2.]),
                        ([2., 3.5], [0.5, 1.]),
                        ([1.], [0.25])]]
    collection_info = [[{'el': 'C', 'index': '1'}],
                       [{'el': 'C', 'index': '1'},
                        {'el': 'C', 'index': '2'}],
                       [{'el': 'C', 'index': '1'},
                        {'el': 'C', 'index': '2'},
                        {'el': 'C', 'index': '2'}]]
    expected_sum = [([1., 2., 3.], [1., 1., 2.],
                     {'el': 'C', 'index': '1'}),
                    ([1., 2., 3., 2., 3.5], [1., 1., 2., 0.5, 1.],
                     {'el': 'C'}),
                    ([1., 2., 3., 2., 3.5, 1.], [1., 1., 2., 0.5, 1., 0.25],
                     {'el': 'C'})]

    @pytest.mark.parametrize('collection_data, collection_info, expected',
                             zip(collection_data, collection_info,
                                 expected_sum))
    def test_sum_all(self, collection_data, collection_info, expected):
        dc = DOSCollection([RawDOSData(*item, info=info)
                            for item, info in zip(collection_data,
                                                  collection_info)])
        summed_dc = dc.sum_all()
        energies, weights, ref_info = expected
        assert np.allclose(summed_dc.get_energies(), energies)
        assert np.allclose(summed_dc.get_weights(), weights)
        assert summed_dc.info == ref_info

    def test_sum_empty(self):
        dc = DOSCollection([])
        with pytest.raises(IndexError):
            dc.sum_all()

    @pytest.mark.parametrize('collection_data, collection_info',
                             zip(collection_data, collection_info))
    def test_total(self, collection_data, collection_info):
        dc = DOSCollection([RawDOSData(*item, info=info)
                            for item, info in zip(collection_data,
                                                  collection_info)])
        summed = dc.sum_all()
        total = dc.total()
        assert np.allclose(summed.get_energies(), total.get_energies())
        assert np.allclose(summed.get_weights(), total.get_weights())
        assert (set(total.info.items()) - set(summed.info.items())
                == {('label', 'Total')})

    select_info = [[{'a': '1', 'b': '1'}, {'a': '2'}],
                   [{'a': '1', 'b': '1'}, {'a': '1', 'b': '2'}],
                   [{'a': '1'}, {'a': '2'}],
                   [{'a': '1', 'b': '1', 'c': '1'},
                    {'a': '1', 'b': '1', 'c': '2'},
                    {'a': '1', 'b': '2', 'c': '3'}]]

    select_query = [{'a': '1'},
                    {'a': '1'},
                    {'a': '0'},
                    {'a': '1', 'b': '1'}]

    select_result = [[{'a': '1', 'b': '1'}],
                     [{'a': '1', 'b': '1'}, {'a': '1', 'b': '2'}],
                     None,
                     [{'a': '1', 'b': '1', 'c': '1'},
                      {'a': '1', 'b': '1', 'c': '2'}]]
    select_not_result = [[{'a': '2'}],
                         None,
                         [{'a': '1'}, {'a': '2'}],
                         [{'a': '1', 'b': '2', 'c': '3'}]]

    sum_by_result = [[{'a': '1', 'b': '1'}, {'a': '2'}],
                     [{'a': '1'}],
                     [{'a': '1'}, {'a': '2'}],
                     [{'a': '1', 'b': '1'}, {'a': '1', 'b': '2', 'c': '3'}]]

    @pytest.mark.parametrize(
        'select_info, select_query, '
        'select_result, select_not_result, sum_by_result',
        zip(select_info, select_query,
            select_result, select_not_result, sum_by_result))
    def test_select(self, select_info, select_query,
                    select_result, select_not_result, sum_by_result):
        dc = DOSCollection([RawDOSData([0.], [0.], info=info)
                            for info in select_info])

        if select_result is None:
            assert dc.select(**select_query)._almost_equals(DOSCollection([]))
        else:
            assert select_result == [data.info for data in
                                     dc.select(**select_query)]

        if select_not_result is None:
            assert (dc.select_not(**select_query)
                    ._almost_equals(DOSCollection([])))
        else:
            assert select_not_result == [data.info for data in
                                         dc.select_not(**select_query)]

        assert sum_by_result == [data.info for data in
                                 dc.sum_by(*sorted(select_query.keys()))]


class TestRawDOSCollection:
    @pytest.fixture()
    def griddos(self):
        energies = np.linspace(1, 10, 7)
        weights = np.sin(energies)
        return GridDOSData(energies, weights, info={'my_key': 'my_value'})

    def test_init(self, griddos):
        with pytest.raises(TypeError):
            RawDOSCollection([griddos])


class TestGridDOSCollection:
    @pytest.fixture()
    def griddos(self):
        energies = np.linspace(1, 10, 7)
        weights = np.sin(energies)
        return GridDOSData(energies, weights, info={'my_key': 'my_value'})

    @pytest.fixture()
    def another_griddos(self):
        energies = np.linspace(1, 10, 7)
        weights = np.cos(energies)
        return GridDOSData(energies, weights, info={'my_key': 'other_value'})

    @pytest.fixture()
    def griddoscollection(self, griddos, another_griddos):
        return GridDOSCollection([griddos, another_griddos])

    def test_init_errors(self, griddos):
        with pytest.raises(TypeError):
            GridDOSCollection([RawDOSData([1.], [1.])])
        with pytest.raises(ValueError):
            energies = np.linspace(1, 10, 7) + 1
            GridDOSCollection([griddos,
                               GridDOSData(energies, np.sin(energies))])
        with pytest.raises(ValueError):
            energies = np.linspace(1, 10, 6)
            GridDOSCollection([griddos,
                               GridDOSData(energies, np.sin(energies))])
        with pytest.raises(ValueError):
            GridDOSCollection([], energies=None)
        with pytest.raises(ValueError):
            GridDOSCollection([griddos], energies=np.linspace(1, 10, 6))

    def test_select(self, griddos, another_griddos):
        gdc = GridDOSCollection([griddos, another_griddos])
        assert (gdc.select(my_key='my_value')
                ._almost_equals(GridDOSCollection([griddos])))
        assert (gdc.select(my_key='not_present')._almost_equals(
            GridDOSCollection([], energies=griddos.get_energies())))
        assert (gdc.select_not(my_key='my_value')
                ._almost_equals(GridDOSCollection([another_griddos])))
        assert (gdc.select(my_key='my_value').select_not(my_key='my_value')
                ._almost_equals(
                    GridDOSCollection([], energies=griddos.get_energies())))

    def test_sequence(self, griddos, another_griddos):
        gdc = GridDOSCollection([griddos, another_griddos])

        for i, (coll_dosdata, dosdata) in enumerate(zip(gdc,
                                                        [griddos,
                                                         another_griddos])):
            assert coll_dosdata._almost_equals(dosdata)
            assert gdc[i]._almost_equals(dosdata)

    def test_slicing(self, griddos, another_griddos):
        gdc = GridDOSCollection([griddos, another_griddos, griddos])

        assert gdc[1:]._almost_equals(
            GridDOSCollection([another_griddos, griddos]))
        assert gdc[:-1]._almost_equals(
            GridDOSCollection([griddos, another_griddos]))

        with pytest.raises(TypeError):
            gdc['string']

    @pytest.mark.parametrize(
        'x, weights, info, error',
        [(np.linspace(1, 10, 12), [np.linspace(4, 1, 12), np.sin(range(12))],
          [{'entry': '1'}, {'entry': '2'}], None),
         (np.linspace(1, 5, 7), [np.sqrt(range(7))], [{'entry': '1'}], None),
         (np.linspace(1, 5, 7), [np.ones((3, 3))], None, IndexError),
         (np.linspace(1, 5, 7), np.array([]).reshape(0, 7), None, IndexError),
         (np.linspace(1, 5, 7), np.ones((2, 6)), None, IndexError)])
    def test_from_data(self, x, weights, info, error):
        if error is not None:
            with pytest.raises(error):
                dc = GridDOSCollection.from_data(x, weights, info=info)
        else:
            dc = GridDOSCollection.from_data(x, weights, info=info)

            for i, dos_data in enumerate(dc):
                assert dos_data.info == info[i]
                assert np.allclose(dos_data.get_energies(), x)
                assert np.allclose(dos_data.get_weights(), weights[i])

    @pytest.mark.usefixtures("figure")
    def test_plot_no_resample(self, griddoscollection, figure):
        ax = figure.add_subplot(111)
        griddoscollection.plot(ax=ax)

        assert np.allclose(ax.get_lines()[0].get_xdata(),
                           griddoscollection[0].get_energies())
        assert np.allclose(ax.get_lines()[1].get_ydata(),
                           griddoscollection[1].get_weights())



import json
import urllib.request
import warnings
from collections import namedtuple
from io import BytesIO, StringIO
from urllib.error import HTTPError, URLError

from ase.io import read

base_url = 'https://pubchem.ncbi.nlm.nih.gov/rest/pug'

PubchemSearch = namedtuple('PubchemSearch', 'search field')


class PubchemData:
    """
    a specialized class for entries from the pubchem database
    """

    def __init__(self, atoms, data):
        self.atoms = atoms
        self.data = data

    def get_atoms(self):
        return self.atoms

    def get_pubchem_data(self):
        return self.data


def search_pubchem_raw(search, field, silent=False, mock_test=False):
    """
    A helper function for searching pubchem.

    Parameters:
        search (str or int):
            the compound you are searching for. This can be either
            a common name, CID, or smiles string depending of the
            `field` you are searching

        field (str):
            the particular field you are searching with. Possible values
            are 'name', 'CID', and 'smiles'.'name' will search common '
            'names,CID will search the Pubchem Chemical Idenitification '
            'Numberswhich can be found on their website and smiles'
            ' searches for compounds with the entered smiles string.

    returns:
        data (str):
            a string containing the raw response from pubchem.
    """
    if mock_test:  # for testing only
        r = BytesIO(test_output)
    else:
        suffix = 'sdf?record_type=3d'

        url = (
            f'{base_url}/{field}/{search!s}/{suffix}'
            if field == 'conformers'
            else f'{base_url}/compound/{field}/{search!s}/{suffix}'
        )
        try:
            r = urllib.request.urlopen(url)
        except HTTPError as e:
            raise ValueError(
                f'the search term {search} could not be found for the field '
                f'{field}'
            ) from e
        except URLError as e:
            raise ValueError(
                'Couldn\'t reach the pubchem servers, check'
                ' your internet connection'
            ) from e

    # check if there are confomers and warn them if there are
    if field != 'conformers' and not silent:
        conformer_ids = available_conformer_search(search, field,
                                                   mock_test=mock_test)
        if len(conformer_ids) > 1:
            warnings.warn(
                f'The structure "{search}" has more than one conformer in '
                'PubChem. By default, the first conformer is returned, please '
                'ensure you are using the structure you intend to or use the '
                '`ase.data.pubchem.pubchem_conformer_search` function'
            )

    return r.read().decode('utf-8')


def parse_pubchem_raw(data):
    """
    a helper function for parsing the returned pubchem entries

    Parameters:
        data (str):
            the raw output from pubchem in string form

    returns:
        atoms (ASE Atoms Object):
            An ASE atoms obejct containing the information from
            pubchem
        pubchem_data (dict):
            a dictionary containing the non-structural information
            from pubchem

    """
    if 'PUBCHEM_COMPOUND_CID' not in data:
        raise Exception('There was a problem with the data returned by '
                        'PubChem')
    f_like = StringIO(data)
    atoms = read(f_like, format='sdf')

    # check if there are confomers and warn them if there are

    # further analyze the text returned from pubchem
    pubchem_data = {}
    other_info = data.split('END\n')[1]
    other_info = other_info.split('$')[0]  # remove the $$$$ at the end
    # the strucuture of this string is > <field>\nentry_info\n
    other_info = other_info.split('> <')  # split into the fields
    for data_field in other_info:
        if data_field == '':
            continue
        field_name, entry_value = data_field.split('>\n')
        # split it into lines and remove the empty lines
        entry_value = entry_value.splitlines()
        entry_value = [a for a in entry_value if a != '']
        if len(entry_value) == 1:
            entry_value = entry_value[0]
        pubchem_data[field_name] = entry_value
    # recover partial charges
    if 'PUBCHEM_MMFF94_PARTIAL_CHARGES' in pubchem_data:
        # the first entry just contains the number of atoms with charges
        charges = pubchem_data['PUBCHEM_MMFF94_PARTIAL_CHARGES'][1:]
        # each subsequent entry contains the index and charge of the atoms
        atom_charges = [0.] * len(atoms)
        for charge in charges:
            i, charge = charge.split()
            # indices start at 1
            atom_charges[int(i) - 1] = float(charge)
        atoms.set_initial_charges(atom_charges)
    return atoms, pubchem_data


def analyze_input(name=None, cid=None, smiles=None, conformer=None,
                  silent=False):
    """
    helper function to translate keyword arguments from intialization
    and searching into the search and field that is being asked for

    Parameters:
        see `ase.data.pubchem.pubchem_search`
    returns:
        search:
            the search term the user has entered
        field:
            the name of the field being asked for

    """
    inputs = [name, cid, smiles, conformer]
    inputs_check = [a is not None for a in [name, cid, smiles, conformer]]
    input_fields = ['name', 'cid', 'smiles', 'conformers']

    if inputs_check.count(True) > 1:
        raise ValueError('Only one search term my be entered a time.'
                         ' Please pass in only one of the following: '
                         'name, cid, smiles, confomer')
    elif inputs_check.count(True) == 1:
        # Figure out which input has been passed in
        index = inputs_check.index(True)
        field = input_fields[index]
        search = inputs[index]
    else:
        raise ValueError('No search was entered.'
                         ' Please pass in only one of the following: '
                         'name, cid, smiles, confomer')

    return PubchemSearch(search, field)


def available_conformer_search(search, field, mock_test=False):
    """
    Helper function to get the conformer IDs. This searches pubchem for
    the conformers of a given structure and returns all the confomer ids
    of a structure.

    Parameters:
        search (str or int):
            the compound you are searching for. This can be either
            a common name, CID, or smiles string depending of the
            `field` you are searching

        field (str):
            the particular field you are searching with. Possible values
            are 'name', 'CID', and 'smiles'.'name' will search common '
            'names,CID will search the Pubchem Chemical Idenitification '
            'Numberswhich can be found on their website and smiles'
            ' searches for compounds with the entered smiles string.

        returns:
            conformers_ids (list):
                a list of the conformer IDs from PubChem, this is different
                than the CID numbers
    """
    suffix = 'conformers/JSON'
    url = f'{base_url}/compound/{field}/{search!s}/{suffix}'
    if mock_test:
        r = BytesIO(test_conformer_output)
    else:
        try:
            r = urllib.request.urlopen(url)
        except HTTPError as e:
            err = ValueError(
                f'the search term {search} could not be found for the field '
                f'{field}'
            )
            raise err from e
        except URLError as e:
            err = ValueError('Couldn\'t reach the pubchem servers, check'
                             ' your internet connection')
            raise err from e
    record = r.read().decode('utf-8')
    record = json.loads(record)
    return record['InformationList']['Information'][0]['ConformerID']


def pubchem_search(*args, mock_test=False, **kwargs):
    """
    Search PubChem for the field and search input on the argument passed in
    returning a PubchemData object. Note that only one argument may be passed
    in at a time.

    Parameters:
        name (str):
            the common name of the compound you're searching for
        cid (str or int):
            the cid of the compound you're searching for
        smiles (str):
            the smiles string of the compound you're searching for
        conformer (str or int):
            the conformer id of the compound you're searching for

    returns:
        result (PubchemData):
            a pubchem data object containing the information on the
            requested entry
    """

    search, field = analyze_input(*args, **kwargs)
    raw_pubchem = search_pubchem_raw(search, field, mock_test=mock_test)
    atoms, data = parse_pubchem_raw(raw_pubchem)
    return PubchemData(atoms, data)


def pubchem_conformer_search(*args, mock_test=False, **kwargs):
    """
    Search PubChem for all the conformers of a given compound.
    Note that only one argument may be passed in at a time.

    Parameters:
        see `ase.data.pubchem.pubchem_search`

    returns:
        conformers (list):
            a list containing the PubchemData objects of all the conformers
            for your search
    """

    search, field = analyze_input(*args, **kwargs)

    conformer_ids = available_conformer_search(search, field,
                                               mock_test=mock_test)
    return [
        pubchem_search(mock_test=mock_test, conformer=id_)
        for id_ in conformer_ids
    ]


def pubchem_atoms_search(*args, **kwargs):
    """
    Search PubChem for the field and search input on the argument passed in
    returning an atoms object.Note that only one argument may be passed
    in at a time.

    Parameters:
        see `ase.data.pubchem.pubchem_search`

    returns:
        atoms (ASE Atoms Object):
            an ASE Atoms object containing the information on the
            requested entry
    """
    return pubchem_search(*args, **kwargs).get_atoms()


def pubchem_atoms_conformer_search(*args, **kwargs):
    """
    Search PubChem for all the conformers of a given compound.
    Note that only one argument may be passed in at a time.

    Parameters:
        see `ase.data.pubchem.pubchem_search`

    returns:
        conformers (list):
            a list containing the atoms objects of all the conformers
            for your search
    """
    conformers = pubchem_conformer_search(*args, **kwargs)
    conformers = [conformer.get_atoms() for conformer in conformers]
    return conformers


test_output = b'222\n  -OEChem-10071914343D\n\n  4  3  0     0  0  0  0  0  0999 V2000\n    0.0000    0.0000    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0\n   -0.4417    0.2906    0.8711 H   0  0  0  0  0  0  0  0  0  0  0  0\n    0.7256    0.6896   -0.1907 H   0  0  0  0  0  0  0  0  0  0  0  0\n    0.4875   -0.8701    0.2089 H   0  0  0  0  0  0  0  0  0  0  0  0\n  1  2  1  0  0  0  0\n  1  3  1  0  0  0  0\n  1  4  1  0  0  0  0\nM  END\n> <PUBCHEM_COMPOUND_CID>\n222\n\n> <PUBCHEM_CONFORMER_RMSD>\n0.4\n\n> <PUBCHEM_CONFORMER_DIVERSEORDER>\n1\n\n> <PUBCHEM_MMFF94_PARTIAL_CHARGES>\n4\n1 -1.08\n2 0.36\n3 0.36\n4 0.36\n\n> <PUBCHEM_EFFECTIVE_ROTOR_COUNT>\n0\n\n> <PUBCHEM_PHARMACOPHORE_FEATURES>\n1\n1 1 cation\n\n> <PUBCHEM_HEAVY_ATOM_COUNT>\n1\n\n> <PUBCHEM_ATOM_DEF_STEREO_COUNT>\n0\n\n> <PUBCHEM_ATOM_UDEF_STEREO_COUNT>\n0\n\n> <PUBCHEM_BOND_DEF_STEREO_COUNT>\n0\n\n> <PUBCHEM_BOND_UDEF_STEREO_COUNT>\n0\n\n> <PUBCHEM_ISOTOPIC_ATOM_COUNT>\n0\n\n> <PUBCHEM_COMPONENT_COUNT>\n1\n\n> <PUBCHEM_CACTVS_TAUTO_COUNT>\n1\n\n> <PUBCHEM_CONFORMER_ID>\n000000DE00000001\n\n> <PUBCHEM_MMFF94_ENERGY>\n0\n\n> <PUBCHEM_FEATURE_SELFOVERLAP>\n5.074\n\n> <PUBCHEM_SHAPE_FINGERPRINT>\n260 1 18410856563934756871\n\n> <PUBCHEM_SHAPE_MULTIPOLES>\n15.6\n0.51\n0.51\n0.51\n0\n0\n0\n0\n0\n0\n0\n0\n0\n0\n\n> <PUBCHEM_SHAPE_SELFOVERLAP>\n14.89\n\n> <PUBCHEM_SHAPE_VOLUME>\n15.6\n\n> <PUBCHEM_COORDINATE_TYPE>\n2\n5\n10\n\n$$$$\n'  # noqa
test_conformer_output = b'{\n  "InformationList": {\n    "Information": [\n      {\n        "CID": 222,\n        "ConformerID": [\n          "000000DE00000001"\n        ]\n      }\n    ]\n  }\n}\n'  # noqa



import numpy as np
from numpy.testing import assert_allclose

from ase import Atoms
from ase.geometry import permute_axes


def test_permute_axes():

    TOL = 1E-10

    rng = np.random.RandomState(0)
    for _ in range(20):
        n = 10
        atoms = Atoms(numbers=[1] * n,
                      scaled_positions=rng.uniform(0, 1, (n, 3)),
                      pbc=rng.randint(0, 2, 3),
                      cell=rng.uniform(-1, 1, (3, 3)))

        permutation = rng.permutation(3)
        permuted = permute_axes(atoms, permutation)
        invperm = np.argsort(permutation)
        original = permute_axes(permuted, invperm)

        assert (original.pbc == atoms.pbc).all()
        assert_allclose(original.cell, atoms.cell, atol=TOL)
        assert_allclose(original.get_positions(),
                        atoms.get_positions(), atol=TOL)
        assert_allclose(atoms.get_positions()[:, permutation],
                        permuted.get_positions(), atol=TOL)

        assert_allclose(atoms.cell.volume, permuted.cell.volume, atol=TOL)
        assert_allclose(atoms.cell.volume, original.cell.volume, atol=TOL)
        assert (permuted.pbc == atoms.pbc[permutation]).all()



pw_keys = {
    "control": [
        "calculation",
        "title",
        "verbosity",
        "restart_mode",
        "wf_collect",
        "nstep",
        "iprint",
        "tstress",
        "tprnfor",
        "dt",
        "outdir",
        "wfcdir",
        "prefix",
        "lkpoint_dir",
        "max_seconds",
        "etot_conv_thr",
        "forc_conv_thr",
        "disk_io",
        "pseudo_dir",
        "tefield",
        "dipfield",
        "lelfield",
        "nberrycyc",
        "lorbm",
        "lberry",
        "gdir",
        "nppstr",
        "gate",
        "twochem",
        "lfcp",
        "trism"
    ],
    "system": [
        "ibrav",
        "celldm",
        "a",
        "b",
        "c",
        "cosab",
        "cosac",
        "cosbc",
        "nat",
        "ntyp",
        "nbnd",
        "nbnd_cond",
        "tot_charge",
        "starting_charge",
        "tot_magnetization",
        "starting_magnetization",
        "ecutwfc",
        "ecutrho",
        "ecutfock",
        "nr1",
        "nr2",
        "nr3",
        "nr1s",
        "nr2s",
        "nr3s",
        "nosym",
        "nosym_evc",
        "noinv",
        "no_t_rev",
        "force_symmorphic",
        "use_all_frac",
        "occupations",
        "one_atom_occupations",
        "starting_spin_angle",
        "degauss_cond",
        "nelec_cond",
        "degauss",
        "smearing",
        "nspin",
        "sic_gamma",
        "pol_type",
        "sic_energy",
        "sci_vb",
        "sci_cb",
        "noncolin",
        "ecfixed",
        "qcutz",
        "q2sigma",
        "input_dft",
        "ace",
        "exx_fraction",
        "screening_parameter",
        "exxdiv_treatment",
        "x_gamma_extrapolation",
        "ecutvcut",
        "nqx1",
        "nqx2",
        "nqx3",
        "localization_thr",
        "hubbard_occ",
        "hubbard_alpha",
        "hubbard_beta",
        "starting_ns_eigenvalue",
        "dmft",
        "dmft_prefix",
        "ensemble_energies",
        "edir",
        "emaxpos",
        "eopreg",
        "eamp",
        "angle1",
        "angle2",
        "lforcet",
        "constrained_magnetization",
        "fixed_magnetization",
        "lambda",
        "report",
        "lspinorb",
        "assume_isolated",
        "esm_bc",
        "esm_w",
        "esm_efield",
        "esm_nfit",
        "lgcscf",
        "gcscf_mu",
        "gcscf_conv_thr",
        "gcscf_beta",
        "vdw_corr",
        "london",
        "london_s6",
        "london_c6",
        "london_rvdw",
        "london_rcut",
        "dftd3_version",
        "dftd3_threebody",
        "ts_vdw_econv_thr",
        "ts_vdw_isolated",
        "xdm",
        "xdm_a1",
        "xdm_a2",
        "space_group",
        "uniqueb",
        "origin_choice",
        "rhombohedral",
        "zgate",
        "relaxz",
        "block",
        "block_1",
        "block_2",
        "block_height",
        "nextffield"
    ],
    "electrons": [
        "electron_maxstep",
        "exx_maxstep",
        "scf_must_converge",
        "conv_thr",
        "adaptive_thr",
        "conv_thr_init",
        "conv_thr_multi",
        "mixing_mode",
        "mixing_beta",
        "mixing_ndim",
        "mixing_fixed_ns",
        "diagonalization",
        "diago_thr_init",
        "diago_cg_maxiter",
        "diago_ppcg_maxiter",
        "diago_david_ndim",
        "diago_rmm_ndim",
        "diago_rmm_conv",
        "diago_gs_nblock",
        "diago_full_acc",
        "efield",
        "efield_cart",
        "efield_phase",
        "startingpot",
        "startingwfc",
        "tqr",
        "real_space"
    ],
    "ions": [
        "ion_positions",
        "ion_velocities",
        "ion_dynamics",
        "pot_extrapolation",
        "wfc_extrapolation",
        "remove_rigid_rot",
        "ion_temperature",
        "tempw",
        "tolp",
        "delta_t",
        "nraise",
        "refold_pos",
        "upscale",
        "bfgs_ndim",
        "trust_radius_max",
        "trust_radius_min",
        "trust_radius_ini",
        "w_1",
        "w_2",
        "fire_alpha_init",
        "fire_falpha",
        "fire_nmin",
        "fire_f_inc",
        "fire_f_dec",
        "fire_dtmax"
    ],
    "cell": [
        "cell_dynamics",
        "press",
        "wmass",
        "cell_factor",
        "press_conv_thr",
        "cell_dofree"
    ],
    "fcp": [
        "fcp_mu",
        "fcp_dynamics",
        "fcp_conv_thr",
        "fcp_ndiis",
        "fcp_mass",
        "fcp_velocity",
        "fcp_temperature",
        "fcp_tempw",
        "fcp_tolp",
        "fcp_delta_t",
        "fcp_nraise",
        "freeze_all_atoms"
    ],
    "rism": [
        "nsolv",
        "closure",
        "tempv",
        "ecutsolv",
        "solute_lj",
        "solute_epsilon",
        "solute_sigma",
        "starting1d",
        "starting3d",
        "smear1d",
        "smear3d",
        "rism1d_maxstep",
        "rism3d_maxstep",
        "rism1d_conv_thr",
        "rism3d_conv_thr",
        "mdiis1d_size",
        "mdiis3d_size",
        "mdiis1d_step",
        "mdiis3d_step",
        "rism1d_bond_width",
        "rism1d_dielectric",
        "rism1d_molesize",
        "rism1d_nproc",
        "rism3d_conv_level",
        "rism3d_planar_average",
        "laue_nfit",
        "laue_expand_right",
        "laue_expand_left",
        "laue_starting_right",
        "laue_starting_left",
        "laue_buffer_right",
        "laue_buffer_left",
        "laue_both_hands",
        "laue_wall",
        "laue_wall_z",
        "laue_wall_rho",
        "laue_wall_epsilon",
        "laue_wall_sigma",
        "laue_wall_lj6"
    ]
}

ph_keys = {
    "inputph": [
        "amass",
        "outdir",
        "prefix",
        "niter_ph",
        "tr2_ph",
        "alpha_mix",
        "nmix_ph",
        "verbosity",
        "reduce_io",
        "max_seconds",
        "dftd3_hess",
        "fildyn",
        "fildrho",
        "fildvscf",
        "epsil",
        "lrpa",
        "lnoloc",
        "trans",
        "lraman",
        "eth_rps",
        "eth_ns",
        "dek",
        "recover",
        "low_directory_check",
        "only_init",
        "qplot",
        "q2d",
        "q_in_band_form",
        "electron_phonon",
        "el_ph_nsigma",
        "el_ph_sigma",
        "ahc_dir",
        "ahc_nbnd",
        "ahc_nbndskip",
        "skip_upperfan",
        "lshift_q",
        "zeu",
        "zue",
        "elop",
        "fpol",
        "ldisp",
        "nogg",
        "asr",
        "ldiag",
        "lqdir",
        "search_sym",
        "nq1",
        "nq2",
        "nq3",
        "nk1",
        "nk2",
        "nk3",
        "k1",
        "k2",
        "k3",
        "diagonalization",
        "read_dns_bare",
        "ldvscf_interpolate",
        "wpot_dir",
        "do_long_range",
        "do_charge_neutral",
        "start_irr",
        "last_irr",
        "nat_todo",
        "modenum",
        "start_q",
        "last_q",
        "dvscf_star",
        "drho_star",
    ]
}

pp_keys = {
    "inputpp": [
        "prefix",
        "outdir",
        "filplot",
        "plot_num",
        "spin_component",
        "spin_component",
        "emin",
        "emax",
        "delta_e",
        "degauss_ldos",
        "use_gauss_ldos",
        "sample_bias",
        "kpoint",
        "kband",
        "lsign",
        "spin_component",
        "emin",
        "emax",
        "spin_component",
        "spin_component",
        "spin_component",
        "spin_component",
    ],
    "plot": [
        "nfile",
        "filepp",
        "weight",
        "iflag",
        "output_format",
        "fileout",
        "interpolation",
        "e1",
        "x0",
        "nx",
        "e1",
        "e2",
        "x0",
        "nx",
        "ny",
        "e1",
        "e2",
        "e3",
        "x0",
        "nx",
        "ny",
        "nz",
        "radius",
        "nx",
        "ny",
    ],
}

matdyn_keys = {
    "input": [
        "flfrc",
        "asr",
        "huang",
        "dos",
        "nk1",
        "nk2",
        "nk3",
        "deltae",
        "ndos",
        "degauss",
        "fldos",
        "flfrq",
        "flvec",
        "fleig",
        "fldyn",
        "at",
        "l1",
        "l2",
        "l3",
        "ntyp",
        "amass",
        "readtau",
        "fltau",
        "la2f",
        "q_in_band_form",
        "q_in_cryst_coord",
        "eigen_similarity",
        "fd",
        "na_ifc",
        "nosym",
        "loto_2d",
        "loto_disable",
        "read_lr",
        "write_frc",
    ]
}


dynmat_keys = {
    "input": [
        "fildyn",
        "q",
        "amass",
        "asr",
        "remove_interaction_blocks",
        "axis",
        "lperm",
        "lplasma",
        "filout",
        "fileig",
        "filmol",
        "filxsf",
        "loto_2d",
        "el_ph_nsig",
        "el_ph_sigma",
    ]
}


q2r_keys = {"input": ["fildyn", "flfrc", "zasr", "loto_2d", "write_lr"]}

dos_keys = {
    "dos": [
        "prefix",
        "outdir",
        "bz_sum",
        "ngauss",
        "degauss",
        "emin",
        "emax",
        "deltae",
        "fildos",
    ]
}


bands_keys = {
    "bands": [
        "prefix",
        "outdir",
        "filband",
        "spin_component",
        "lsigma",
        "lp",
        "filp",
        "lsym",
        "no_overlap",
        "plot_2d",
        "firstk",
        "lastk",
    ]
}


band_interpolation_keys = {
    "interpolation": [
        "method",
        "miller_max",
        "check_periodicity",
        "p_metric",
        "scale_sphere",
    ]
}


projwfc_keys = {
    "projwfc": [
        "prefix",
        "outdir",
        "ngauss",
        "degauss",
        "emin",
        "emax",
        "deltae",
        "lsym",
        "diag_basis",
        "pawproj",
        "filpdos",
        "filproj",
        "lwrite_overlaps",
        "lbinary_data",
        "kresolveddos",
        "tdosinboxes",
        "n_proj_boxes",
        "irmin",
        "irmax",
        "plotboxes",
    ]
}


molecularpdos_keys = {
    "inputmopdos": [
        "xmlfile_full",
        "xmlfile_part",
        "i_atmwfc_beg_full",
        "i_atmwfc_end_full",
        "i_atmwfc_beg_part",
        "i_atmwfc_end_part",
        "i_bnd_beg_full",
        "i_bnd_end_full",
        "i_bnd_beg_part",
        "i_bnd_end_part",
        "fileout",
        "ngauss",
        "degauss",
        "emin",
        "emax",
        "deltae",
        "kresolveddos",
    ]
}


importexport_binary_keys = {
    "inputpp": [
        "prefix",
        "outdir",
        "direction",
        "newoutdir"]}


oscdft_pp_keys = {"oscdft_pp_namelist": ["prefix", "outdir"]}

kcw_keys = {
    "control": [
        "prefix",
        "outdir",
        "calculation",
        "kcw_iverbosity",
        "kcw_at_ks",
        "read_unitary_matrix",
        "spread_thr",
        "homo_only",
        "l_vcut",
        "assume_isolated",
        "spin_component",
        "mp1",
        "mp2",
        "mp3",
        "lrpa",
    ],
    "wannier": [
        "seedname",
        "num_wann_occ",
        "num_wann_emp",
        "have_empty",
        "has_disentangle",
        "check_ks",
    ],
    "screen": ["niter", "nmix", "tr2", "i_orb", "eps_inf", "check_spread"],
    "ham": ["do_bands", "use_ws_distance", "write_hr", "on_site_only"],
}


cppp_keys = {
    "inputpp": [
        "prefix",
        "fileout",
        "output",
        "outdir",
        "lcharge",
        "lforces",
        "ldynamics",
        "lpdb",
        "lrotation",
        "np1",
        "np2",
        "np3",
        "nframes",
        "ndr",
        "atomic_number",
    ]
}


ppacf_keys = {
    "ppacf": [
        "prefix",
        "outdir",
        "n_lambda",
        "lplot",
        "ltks",
        "lfock",
        "use_ace",
        "code_num",
        "vdw_analysis",
    ]
}


all_currents_keys = {
    "energy_current": [
        "delta_t",
        "file_output",
        "trajdir",
        "vel_input_units",
        "eta",
        "n_max",
        "first_step",
        "last_step",
        "step_mul",
        "step_rem",
        "ethr_small_step",
        "ethr_big_step",
        "restart",
        "subtract_cm_vel",
        "add_i_current_b",
        "save_dvpsi",
        "re_init_wfc_1",
        "re_init_wfc_2",
        "re_init_wfc_3",
        "three_point_derivative",
        "n_repeat_every_step",
        "n_workers",
        "worker_id",
        "continue_not_converged",
    ]
}


turbo_lanczos_keys = {
    "lr_input": [
        "prefix",
        "outdir",
        "wfcdir",
        "restart",
        "restart_step",
        "lr_verbosity",
        "disk_io",
    ],
    "lr_control": [
        "itermax",
        "ipol",
        "n_ipol",
        "ltammd",
        "no_hxc",
        "lrpa",
        "scissor",
        "charge_response",
        "pseudo_hermitian",
        "d0psi_rs",
        "lshift_d0psi",
    ],
    "lr_post": [
        "omeg",
        "epsil",
        "beta_gamma_z_prefix",
        "w_t_npol",
        "plot_type"],
}


turbo_spectrum_keys = {
    "lr_input": [
        "prefix",
        "outdir",
        "verbosity",
        "itermax0",
        "itermax",
        "extrapolation",
        "epsil",
        "units",
        "start",
        "end",
        "increment",
        "ipol",
        "eels",
        "magnons",
        "td",
        "eign_file",
    ]
}


turbo_davidson_keys = {
    "lr_input": [
        "prefix",
        "outdir",
        "wfcdir",
        "max_seconds",
        "restart",
        "lr_verbosity",
        "disk_io",
    ],
    "lr_dav": [
        "num_eign",
        "num_init",
        "if_random_init",
        "num_basis_max",
        "residue_conv_thr",
        "precondition",
        "single_pole",
        "if_dft_spectrum",
        "reference",
        "broadening",
        "start",
        "finish",
        "step",
        "p_nbnd_occ",
        "p_nbnd_virt",
        "poor_of_ram",
        "poor_of_ram2",
        "max_iter",
        "no_hxc",
        "pseudo_hermitian",
        "ltammd",
        "lplot_drho",
        "d0psi_rs",
        "lshift_d0psi",
    ],
}


turbo_magnon_keys = {
    "lr_input": [
        "prefix",
        "outdir",
        "restart",
        "restart_step",
        "lr_verbosity",
        "disk_io",
    ],
    "lr_control": [
        "itermax",
        "pseudo_hermitian",
        "approximation",
        "ipol",
        "q1",
        "q2",
        "q3",
    ],
}


turbo_eels_keys = {
    "lr_input": [
        "prefix",
        "outdir",
        "restart",
        "restart_step",
        "lr_verbosity",
        "disk_io",
    ],
    "lr_control": [
        "approximation",
        "q1",
        "q2",
        "q3",
        "calculator",
        "itermax",
        "pseudo_hermitian",
        "alpha_mix(i",
        "epsil",
        "units",
        "start",
        "end",
        "increment",
        "ethr_nscf",
    ],
}


oscdft_et_keys = {
    "oscdft_et_namelist": [
        "initial_prefix",
        "final_prefix",
        "initial_dir",
        "final_dir",
        "print_matrix",
        "print_eigvect",
        "print_debug",
    ]
}


pprism_keys = {
    "inputpp": ["prefix", "outdir", "filplot", "lpunch"],
    "plot": [
        "iflag",
        "output_format",
        "fileout",
        "interpolation",
        "x0",
        "nx",
        "lebedev",
        "e1",
        "x0",
        "nx",
        "e1",
        "e2",
        "x0",
        "nx",
        "ny",
        "e1",
        "e2",
        "e3",
        "x0",
        "nx",
        "ny",
        "nz",
        "radius",
        "nx",
        "ny",
    ],
}


pwcond_keys = {
    "inputcond": [
        "outdir",
        "prefixt",
        "prefixl",
        "prefixs",
        "prefixr",
        "tran_prefix",
        "max_seconds",
        "recover",
        "band_file",
        "tran_file",
        "save_file",
        "fil_loc",
        "lwrite_cond",
        "loop_ek",
        "lread_cond",
        "lwrite_loc",
        "lread_loc",
        "ikind",
        "iofspin",
        "tk_plot",
        "llocal",
        "bdl",
        "bds",
        "bdr",
        "nz1",
        "energy0",
        "denergy",
        "nenergy",
        "start_e",
        "last_e",
        "start_k",
        "last_k",
        "ecut2d",
        "ewind",
        "epsproj",
        "orbj_in",
        "orbj_fin",
    ]
}


pw2bgw_keys = {
    "input_pw2bgw": [
        "prefix",
        "outdir",
        "real_or_complex",
        "symm_type",
        "wfng_flag",
        "wfng_file",
        "wfng_kgrid",
        "wfng_nk1",
        "wfng_nk2",
        "wfng_nk3",
        "wfng_dk1",
        "wfng_dk2",
        "wfng_dk3",
        "wfng_occupation",
        "wfng_nvmin",
        "wfng_nvmax",
        "rhog_flag",
        "rhog_file",
        "rhog_nvmin",
        "rhog_nvmax",
        "vxcg_flag",
        "vxcg_file",
        "vxc0_flag",
        "vxc0_file",
        "vxc_flag",
        "vxc_file",
        "vxc_integral",
        "vxc_diag_nmin",
        "vxc_diag_nmax",
        "vxc_offdiag_nmin",
        "vxc_offdiag_nmax",
        "vxc_zero_rho_core",
        "vscg_flag",
        "vscg_file",
        "vkbg_flag",
        "vkbg_file",
    ]
}


bgw2pw_keys = {
    "input_bgw2pw": [
        "prefix",
        "outdir",
        "real_or_complex",
        "wfng_flag",
        "wfng_file",
        "wfng_nband",
        "rhog_flag",
        "rhog_file",
    ]
}


hp_keys = {
    "inputhp": [
        "prefix",
        "outdir",
        "iverbosity",
        "max_seconds",
        "nq1",
        "nq2",
        "nq3",
        "skip_equivalence_q",
        "determine_num_pert_only",
        "determine_q_mesh_only",
        "find_atpert",
        "docc_thr",
        "skip_type",
        "equiv_type",
        "perturb_only_atom",
        "start_q",
        "last_q",
        "sum_pertq",
        "compute_hp",
        "conv_thr_chi",
        "thresh_init",
        "ethr_nscf",
        "niter_max",
        "alpha_mix(i",
        "nmix",
        "num_neigh",
        "lmin",
        "rmax",
        "dist_thr",
    ]
}


cp_keys = {
    "control": [
        "calculation",
        "title",
        "verbosity",
        "isave",
        "restart_mode",
        "nstep",
        "iprint",
        "tstress",
        "tprnfor",
        "dt",
        "outdir",
        "saverho",
        "prefix",
        "ndr",
        "ndw",
        "tabps",
        "max_seconds",
        "etot_conv_thr",
        "forc_conv_thr",
        "ekin_conv_thr",
        "disk_io",
        "memory",
        "pseudo_dir",
        "tefield",
    ],
    "system": [
        "ibrav",
        "celldm",
        "a",
        "b",
        "c",
        "cosab",
        "cosac",
        "cosbc",
        "nat",
        "ntyp",
        "nbnd",
        "tot_charge",
        "tot_magnetization",
        "ecutwfc",
        "ecutrho",
        "nr1",
        "nr2",
        "nr3",
        "nr1s",
        "nr2s",
        "nr3s",
        "nr1b",
        "nr2b",
        "nr3b",
        "occupations",
        "degauss",
        "smearing",
        "nspin",
        "ecfixed",
        "qcutz",
        "q2sigma",
        "input_dft",
        "exx_fraction",
        "lda_plus_u",
        "hubbard_u",
        "vdw_corr",
        "london_s6",
        "london_rcut",
        "ts_vdw",
        "ts_vdw_econv_thr",
        "ts_vdw_isolated",
        "assume_isolated",
    ],
    "electrons": [
        "electron_maxstep",
        "electron_dynamics",
        "conv_thr",
        "niter_cg_restart",
        "efield",
        "epol",
        "emass",
        "emass_cutoff",
        "orthogonalization",
        "ortho_eps",
        "ortho_max",
        "ortho_para",
        "electron_damping",
        "electron_velocities",
        "electron_temperature",
        "ekincw",
        "fnosee",
        "startingwfc",
        "tcg",
        "maxiter",
        "passop",
        "pre_state",
        "n_inner",
        "niter_cold_restart",
        "lambda_cold",
        "grease",
        "ampre",
    ],
    "ions": [
        "ion_dynamics",
        "ion_positions",
        "ion_velocities",
        "ion_damping",
        "ion_radius",
        "iesr",
        "ion_nstepe",
        "remove_rigid_rot",
        "ion_temperature",
        "tempw",
        "fnosep",
        "tolp",
        "nhpcl",
        "nhptyp",
        "nhgrp",
        "fnhscl",
        "ndega",
        "tranp",
        "amprp",
        "greasp",
    ],
    "cell": [
        "cell_parameters",
        "cell_dynamics",
        "cell_velocities",
        "cell_damping",
        "press",
        "wmass",
        "cell_factor",
        "cell_temperature",
        "temph",
        "fnoseh",
        "greash",
        "cell_dofree",
    ],
    "press_ai": [
        "abivol",
        "abisur",
        "p_ext",
        "pvar",
        "p_in",
        "p_fin",
        "surf_t",
        "rho_thr",
        "dthr",
    ],
    "wannier": [
        "wf_efield",
        "wf_switch",
        "sw_len",
        "efx0",
        "efy0",
        "efz0",
        "efx1",
        "efy1",
        "efz1",
        "wfsd",
        "wfdt",
        "maxwfdt",
        "nit",
        "nsd",
        "wf_q",
        "wf_friction",
        "nsteps",
        "tolw",
        "adapt",
        "calwf",
        "nwf",
        "wffort",
        "writev",
        "exx_neigh",
        "exx_dis_cutoff",
        "exx_poisson_eps",
        "exx_use_cube_domain",
        "exx_ps_rcut_self",
        "exx_ps_rcut_pair",
        "exx_me_rcut_self",
        "exx_me_rcut_pair",
    ],
}


ld1_keys = {
    "input": [
        "title",
        "zed",
        "atom",
        "xmin",
        "dx",
        "rmax",
        "beta",
        "tr2",
        "iswitch",
        "nld",
        "rlderiv",
        "eminld",
        "emaxld",
        "deld",
        "rpwe",
        "rel",
        "lsmall",
        "max_out_wfc",
        "noscf",
        "lsd",
        "dft",
        "latt",
        "isic",
        "rytoev_fact",
        "cau_fact",
        "vdw",
        "prefix",
        "verbosity",
        "file_charge",
        "config",
        "relpert",
        "rel_dist",
        "write_coulomb",
    ],
    "inputp": [
        "zval",
        "pseudotype",
        "file_pseudopw",
        "file_recon",
        "lloc",
        "rcloc",
        "nlcc",
        "new_core_ps",
        "rcore",
        "tm",
        "rho0",
        "lpaw",
        "which_augfun",
        "rmatch_augfun",
        "rmatch_augfun_nc",
        "lsave_wfc",
        "lgipaw_reconstruction",
        "use_paw_as_gipaw",
        "author",
        "file_chi",
        "file_beta",
        "file_qvan",
        "file_screen",
        "file_core",
        "file_wfcaegen",
        "file_wfcncgen",
        "file_wfcusgen",
    ],
    "test": [
        "nconf",
        "file_pseudo",
        "ecutmin",
        "ecutmax",
        "decut",
        "rm",
        "configts",
        "lsdts",
        "frozen_core",
        "rcutv",
    ],
}


d3hess_keys = {"input": ["prefix", "outdir", "filhess", "step"]}

neb_keys = {
    "path": [
        "string_method",
        "restart_mode",
        "nstep_path",
        "num_of_images",
        "opt_scheme",
        "ci_scheme",
        "first_last_opt",
        "minimum_image",
        "temp_req",
        "ds",
        "k_max",
        "k_min",
        "path_thr",
        "use_masses",
        "use_freezing",
        "lfcp",
        "fcp_mu",
        "fcp_thr",
        "fcp_scheme",
    ]
}

postahc_keys = {
    "input": [
        "ahc_dir",
        "nk",
        "nbnd",
        "nat",
        "nq",
        "ahc_nbnd",
        "ahc_nbndskip",
        "flvec",
        "eta",
        "temp_kelvin",
        "efermi",
        "amass_amu",
        "skip_upperfan",
        "skip_dw"
    ]
}

dvscf_q2r = {
    "input": [
        "prefix",
        "outdir",
        "fildyn",
        "fildvscf",
        "wpot_dir",
        "do_long_range",
        "do_charge_neutral",
        "verbosity"
    ]
}

epw_keys = {
    "inputepw": [
        "adapt_ethrdg_plrn",
        "a2f",
        "amass",
        "asr_typ",
        "assume_metal",
        "band_plot",
        "bands_skipped",
        "bfieldx",
        "bfieldy",
        "bfieldz",
        "bnd_cum",
        "broyden_beta",
        "broyden_ndim",
        "cal_psir_plrn",
        "carrier",
        "conv_thr_iaxis",
        "conv_thr_plrn",
        "conv_thr_racon",
        "conv_thr_raxis",
        "cumulant",
        "degaussq",
        "degaussw",
        "delta_approx",
        "delta_qsmear",
        "delta_smear",
        "dvscf_dir",
        "do_CHBB",
        "efermi_read",
        "eig_read",
        "elecselfen",
        "eliashberg",
        "elph",
        "ep_coupling",
        "epbwrite",
        "epbread",
        "epexst",
        "ephwrite",
        "epmatkqread",
        "eps_acustic",
        "epsiHEG",
        "epwread",
        "epwwrite",
        "etf_mem",
        "ethrdg_plrn",
        "fermi_diff",
        "fermi_energy",
        "fermi_plot",
        "fila2f",
        "fildvscf",
        "filkf",
        "filqf",
        "filukk",
        "filukq",
        "fixsym",
        "fsthick",
        "gap_edge",
        "gb_scattering",
        "gb_only",
        "gb_size",
        "imag_read",
        "init_ethrdg_plrn",
        "init_k0_plrn",
        "init_ntau_plrn",
        "init_plrn",
        "init_sigma_plrn",
        "interp_Ank_plrn",
        "interp_Bqu_plrn",
        "int_mob",
        "io_lvl_plrn",
        "iterative_bte",
        "iverbosity",
        "ii_g",
        "ii_scattering",
        "ii_only",
        "ii_lscreen",
        "ii_partion",
        "ii_charge",
        "ii_n",
        "ii_eda",
        "kerread",
        "kerwrite",
        "kmaps",
        "lacon",
        "laniso",
        "lifc",
        "limag",
        "lindabs",
        "liso",
        "longrange",
        "lpade",
        "lphase",
        "lpolar",
        "lreal",
        "lscreen",
        "lunif",
        "loptabs",
        "len_mesh",
        "max_memlt",
        "meff",
        "mob_maxiter",
        "mp_mesh_k",
        "mp_mesh_q",
        "muc",
        "meshnum",
        "nbndsub",
        "ncarrier",
        "nc",
        "nel",
        "nest_fn",
        "nethrdg_plrn",
        "ngaussw",
        "niter_plrn",
        "nk1",
        "nk2",
        "nk3",
        "nkf1",
        "nkf2",
        "nkf3",
        "nq1",
        "nq2",
        "nq3",
        "nqf1",
        "nqf2",
        "nqf3",
        "npade",
        "nqsmear",
        "nqstep",
        "n_r",
        "nsiter",
        "nsmear",
        "nstemp",
        "nswi",
        "nswc",
        "nswfc",
        "nw",
        "nw_specfun",
        "nq_init",
        "omegamax",
        "omegamin",
        "omegastep",
        "phonselfen",
        "plselfen",
        "plrn",
        "prefix",
        "prtgkk",
        "pwc",
        "QD_bin",
        "QD_min",
        "rand_nq",
        "rand_nk",
        "rand_q",
        "rand_k",
        "restart",
        "restart_filq",
        "restart_plrn",
        "restart_step",
        "scell_mat",
        "scell_mat_plrn",
        "scr_typ",
        "scatread",
        "scattering",
        "scattering_serta",
        "scattering_0rta",
        "scissor",
        "selecqread",
        "smear_rpa",
        "specfun_el",
        "specfun_ph",
        "specfun_pl",
        "system_2d",
        "shortrange",
        "step_wf_grid_plrn",
        "start_mesh",
        "temps",
        "tc_linear",
        "tc_linear_solver",
        "type_plrn",
        "vme",
        "wannierize",
        "wepexst",
        "wmax",
        "wmax_specfun",
        "wmin",
        "wmin_specfun",
        "wscut",
        "wsfc",
        "auto_projections",
        "dis_froz_min",
        "dis_froz_max",
        "iprint",
        "num_iter",
        "proj",
        "reduce_unk",
        "scdm_entanglement",
        "scdm_mu",
        "scdm_proj",
        "scdm_sigma",
        "wannier_plot",
        "wannier_plot_list",
        "wannier_plot_radius",
        "wannier_plot_scale",
        "wannier_plot_supercell",
        "wdata"
    ]
}

ALL_KEYS = {
    "pw": pw_keys,
    "ph": ph_keys,
    "phcg": ph_keys,
    "pp": pp_keys,
    "matdyn": matdyn_keys,
    "dynmat": dynmat_keys,
    "q2r": q2r_keys,
    "dos": dos_keys,
    "bands": bands_keys,
    "bands_interpolation": band_interpolation_keys,
    "projwfc": projwfc_keys,
    "molecularpdos": molecularpdos_keys,
    "importexport_binary": importexport_binary_keys,
    "oscdft_pp": oscdft_pp_keys,
    "kcw": kcw_keys,
    "cppp": cppp_keys,
    "ppacf": ppacf_keys,
    "all_currents": all_currents_keys,
    "turbo_lanczos": turbo_lanczos_keys,
    "turbo_spectrum": turbo_spectrum_keys,
    "turbo_davidson": turbo_davidson_keys,
    "turbo_magnon": turbo_magnon_keys,
    "turbo_eels": turbo_eels_keys,
    "oscdft_et": oscdft_et_keys,
    "pprism": pprism_keys,
    "pwcond": pwcond_keys,
    "pw2bgw": pw2bgw_keys,
    "bgw2pw": bgw2pw_keys,
    "hp": hp_keys,
    "cp": cp_keys,
    "ld1": ld1_keys,
    "d3hess": d3hess_keys,
    "neb": neb_keys,
    "postahc": postahc_keys,
    "dvscf_q2r": dvscf_q2r,
    "epw": epw_keys
}



import numpy as np
import pytest

from ase.build import molecule
from ase.calculators.ff import ForceField
from ase.optimize.precon import FF
from ase.optimize.precon.lbfgs import PreconLBFGS
from ase.optimize.precon.neighbors import get_neighbours
from ase.utils.ff import Angle, Dihedral, Morse, VdW


@pytest.fixture(scope='module')
def atoms0():
    a = molecule('C60')
    a.set_cell(50.0 * np.identity(3))
    return a


@pytest.fixture(scope='module')
def forcefield_params(atoms0):
    # force field parameters for fulleren, Z. Berkai at al.
    # Energy Procedia, 74, 2015, 59-64
    a = atoms0
    cutoff = 1.5
    morse_D = 6.1322
    morse_alpha = 1.8502
    morse_r0 = 1.4322
    angle_k = 10.0
    angle_a0 = np.deg2rad(120.0)
    dihedral_k = 0.346
    vdw_epsilonij = 0.0115
    vdw_rminij = 3.4681

    neighbor_list = [[] for _ in range(len(a))]
    vdw_list = np.ones((len(a), len(a)), dtype=bool)
    morses = []
    angles = []
    dihedrals = []
    vdws = []

    # create neighbor list
    i_list, j_list, _d_list, _fixed_atoms = get_neighbours(
        atoms=a, r_cut=cutoff)

    for i, j in zip(i_list, j_list):
        neighbor_list[i].append(j)
    for neighbor in neighbor_list:
        neighbor.sort()

    # create lists of morse, bending and torsion interactions
    for i in range(len(a)):
        for jj in range(len(neighbor_list[i])):
            j = neighbor_list[i][jj]
            if j > i:
                morses.append(Morse(atomi=i, atomj=j, D=morse_D,
                                    alpha=morse_alpha, r0=morse_r0))
            vdw_list[i, j] = vdw_list[j, i] = False
            for kk in range(jj + 1, len(neighbor_list[i])):
                k = neighbor_list[i][kk]
                angles.append(Angle(atomi=j, atomj=i, atomk=k, k=angle_k,
                                    a0=angle_a0, cos=True))
                vdw_list[j, k] = vdw_list[k, j] = False
                for ll in range(kk + 1, len(neighbor_list[i])):
                    L = neighbor_list[i][ll]
                    dihedrals.append(Dihedral(atomi=j, atomj=i, atomk=k,
                                              atoml=L,
                                              k=dihedral_k))

    # create list of van der Waals interactions
    for i in range(len(a)):
        for j in range(i + 1, len(a)):
            if vdw_list[i, j]:
                vdws.append(VdW(atomi=i, atomj=j, epsilonij=vdw_epsilonij,
                                rminij=vdw_rminij))

    return dict(morses=morses, angles=angles, dihedrals=dihedrals, vdws=vdws)


@pytest.fixture()
def calc(forcefield_params):
    return ForceField(**forcefield_params)


@pytest.fixture()
def atoms(atoms0, calc):
    atoms = atoms0.copy()
    atoms.calc = calc
    atoms.rattle(0.05)
    return atoms


ref_energy = 17.238525

# @pytest.mark.skip('FAILS WITH PYAMG')


@pytest.mark.optimize()
@pytest.mark.slow()
def test_opt_with_precon(atoms, forcefield_params):
    kw = dict(forcefield_params)
    kw.pop('vdws')
    precon = FF(**kw)
    opt = PreconLBFGS(atoms, use_armijo=True, precon=precon)
    opt.run(fmax=0.1)
    e = atoms.get_potential_energy()
    assert abs(e - ref_energy) < 0.01



"""Velocity Verlet."""
from ase.md.md import MolecularDynamics


class VelocityVerlet(MolecularDynamics):
    """MD with NVE ensemble and velocity Verlet time integration."""

    def step(self, forces=None):

        atoms = self.atoms

        if forces is None:
            forces = atoms.get_forces(md=True)

        p = atoms.get_momenta()
        p += 0.5 * self.dt * forces
        masses = atoms.get_masses()[:, None]
        r = atoms.get_positions()

        # if we have constraints then this will do the first part of the
        # RATTLE algorithm:
        atoms.set_positions(r + self.dt * p / masses)
        if atoms.constraints:
            p = (atoms.get_positions() - r) * masses / self.dt

        # We need to store the momenta on the atoms before calculating
        # the forces, as in a parallel Asap calculation atoms may
        # migrate during force calculations, and the momenta need to
        # migrate along with the atoms.
        atoms.set_momenta(p, apply_constraint=False)

        forces = atoms.get_forces(md=True)

        # Second part of RATTLE will be done here:
        atoms.set_momenta(atoms.get_momenta() + 0.5 * self.dt * forces)
        return forces



"""Check that our tab-completion script has been updated."""
from ase.cli.completion import path, update
from ase.cli.main import commands


def test_complete():
    update(path, commands, test=True)



import numpy as np
import pytest
from numpy.random import RandomState

from ase.build import bulk
from ase.data import atomic_numbers
from ase.md.velocitydistribution import PhononHarmonics
from ase.optimize import FIRE
from ase.phonons import Phonons


@pytest.mark.slow()
def test_phonon_md_init(asap3, testdir):
    # Tests the phonon-based perturbation and velocity distribution
    # for thermal equilibration in MD.

    EMT = asap3.EMT

    rng = RandomState(17)

    atoms = bulk('Pd')
    atoms *= (3, 3, 3)
    avail = [atomic_numbers[sym]
             for sym in ['Ni', 'Cu', 'Pd', 'Ag', 'Pt', 'Au']]
    atoms.numbers[:] = rng.choice(avail, size=len(atoms))
    atoms.calc = EMT()

    with FIRE(atoms, trajectory='relax.traj') as opt:
        opt.run(fmax=0.001)
    positions0 = atoms.positions.copy()

    phonons = Phonons(atoms, EMT(), supercell=(1, 1, 1), delta=0.05)

    try:
        phonons.run()
        phonons.read()  # Why all this boilerplate?
    finally:
        phonons.clean()
    matrices = phonons.get_force_constant()

    K = matrices[0]
    T = 300

    atoms.calc = EMT()
    Epotref = atoms.get_potential_energy()

    temps = []
    Epots = []
    Ekins = []
    Etots = []

    for i in range(24):
        PhononHarmonics(atoms, K, temperature_K=T, quantum=True,
                        rng=np.random.RandomState(888 + i))

        Epot = atoms.get_potential_energy() - Epotref
        Ekin = atoms.get_kinetic_energy()
        Ekins.append(Ekin)
        Epots.append(Epot)
        Etots.append(Ekin + Epot)
        temps.append(atoms.get_temperature())

        atoms.positions[:] = positions0

        # The commented code would produce displacements/velocities
        # resolved over phonon modes if we borrow some expressions
        # from the function.  Each mode should contribute on average
        # equally to both Epot and Ekin/temperature
        #
        # atoms1.calc = EMT()
        # atoms1 = atoms.copy()
        # v_ac = np.zeros_like(positions0)
        # D_acs, V_acs = ...
        # for s in range(V_acs.shape[2]):
        #     atoms1.positions += D_acs[:, :, s]
        #     v_ac += V_acs[:, :, s]
        #     atoms1.set_velocities(v_ac)
        #     X1.append(atoms1.get_potential_energy() - Epotref)
        #     X2.append(atoms1.get_kinetic_energy())

        print('energies', Epot, Ekin, Epot + Ekin)

    Epotmean = np.mean(Epots)
    Ekinmean = np.mean(Ekins)
    Tmean = np.mean(temps)
    Terr = abs(Tmean - T)
    relative_imbalance = abs(Epotmean - Ekinmean) / (Epotmean + Ekinmean)

    print('epotmean', Epotmean)
    print('ekinmean', Ekinmean)
    print('rel imbalance', relative_imbalance)
    print('Tmean', Tmean, 'Tref', T, 'err', Terr)

    assert Terr < 0.1 * T, Terr  # error in Kelvin for instantaneous velocity
    # Epot == Ekin give or take 2 %:
    assert relative_imbalance < 0.1, relative_imbalance



"""
Tests of the plugin functionality for defining IO formats
outside of the ase package
"""
import copy
import io
from importlib.metadata import EntryPoint

import pytest

from ase.build import bulk
from ase.io import formats, read, write
from ase.io.formats import define_external_io_format
from ase.utils.plugins import ExternalIOFormat


@pytest.fixture(autouse=True)
def reset_ioformats_after_test():
    ioformats_before = copy.deepcopy(formats.ioformats)
    try:
        yield
    finally:
        formats.ioformats = ioformats_before
        formats.all_formats = ioformats_before


VALID_IO_FORMAT = ExternalIOFormat(
    desc='Test IO format',
    code='1F',
    module='ase.test.fio.test_external_io_formats'
)


# These are dummy functions for reading and writing the dummy io format
def read_dummy(file):
    return "Atoms dummy"


def write_dummy(file, atoms):
    file.write("dummy output")


def test_external_ioformat_valid(tmp_path):
    """
    Test of the external io format utility correctly
    registering a valid external entry format
    """

    test_entry_point = EntryPoint(
        name='dummy',
        value='ase.test.fio.test_external_io_formats:VALID_IO_FORMAT',
        group='ase.ioformats')

    define_external_io_format(test_entry_point)

    assert 'dummy' in formats.ioformats

    TEST_FILE = """
        THIS IS A DUMMY
    """
    assert read(io.StringIO(TEST_FILE), format='dummy') == 'Atoms dummy'

    atom = bulk('Ti')
    write(tmp_path / 'dummy_output', atom, format='dummy')
    with open(tmp_path / 'dummy_output') as file:
        assert file.read() == 'dummy output'


def test_external_ioformat_already_existing():
    """
    Test of the external io format utility correctly
    refusing to register an IOformat that is already present
    """

    test_entry_point = EntryPoint(
        name='xyz',
        value='ase.test.fio.test_external_io_formats:VALID_IO_FORMAT',
        group='ase.ioformats')

    with pytest.raises(ValueError, match='Format xyz already defined'):
        define_external_io_format(test_entry_point)

    assert 'xyz' in formats.ioformats
    assert formats.ioformats['xyz'].description != 'Test IO format'


# Io format not specified with the required namedtuple
INVALID_IO_FORMAT = {
    'desc': 'Test IO format',
    'code': '1F',
    'module': 'ase.test.fio.test_external_io_formats'
}


def test_external_ioformat_wrong_type():
    """
    Test of the external io format utility correctly
    refusing to register an IOformat that is not specified using the
    namedtuple
    """

    test_entry_point = EntryPoint(
        name='dummy',
        value='ase.test.fio.test_external_io_formats:INVALID_IO_FORMAT',
        group='ase.ioformats')

    with pytest.raises(TypeError,
                       match='Wrong type for registering external IO formats'):
        define_external_io_format(test_entry_point)

    assert 'dummy' not in formats.ioformats


def test_external_ioformat_import_error():
    """
    Test of the external io format utility correctly
    refusing to register an IOformat that is already present
    """

    test_entry_point = EntryPoint(
        name='dummy',
        value='ase.test.fio.test_external_io_formats:NOT_EXISTING',
        group='ase.ioformats')

    with pytest.raises(AttributeError):
        define_external_io_format(test_entry_point)

    assert 'dummy' not in formats.ioformats
